using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.Security;
using System.Security.Permissions;
using System.Reflection;
using System.Net;

namespace tine
{
  public static class Constants
  {
    public const int FEC_NAME_SIZE = 16;
    public const int EXPORT_NAME_SIZE = 32;
    public const int EQM_NAME_SIZE = 8;
    public const int EQM_NAME_SHORTSIZE = 6;
    public const int CONTEXT_NAME_SIZE = 32;
    public const int STATUS_SIZE = 192;
    public const int SUBSYSTEM_NAME_SIZE = 16;
    public const int TAG_NAME_SIZE = 16;
    public const int FIELD_NAME_SIZE = 16;
    public const int USERNAME_SIZE = 16;
    public const int DEVICE_NAME_SIZE = 64;
    public const int DEVICE_NAME_MAXSIZE = 1024;
    public const int DEVICE_REDIR_SIZE = STATUS_SIZE;
    public const int PROPERTY_NAME_SIZE = 64;
    public const int PROPERTY_DESC_SIZE = 64;
    public const int PROPERTY_REDIR_SIZE = STATUS_SIZE;
    public const int UNITS_SIZE = 16;
    public const int ADDR_SIZE = 16;
    public const int ALARM_TAG_SIZE = 32;
    public const int ALARM_TEXT_SIZE = 64;
    public const int ALARM_TEXT_LONGSIZE = 128;
  }
  /// <summary>
  /// Available Link Transfer Modes
  /// </summary>
  public static class Modes
  {
    public const UInt16 CM_NULL = 0x00;
    public const UInt16 CM_CANCEL = 0x00;
    public const UInt16 CM_SINGLE = 0x01;
    public const UInt16 CM_REFRESH = 0x02;
    public const UInt16 CM_DATACHANGE = 0x02;
    public const UInt16 CM_POLL = 0x03;
    public const UInt16 CM_TIMER = 0x03;
    public const UInt16 CM_REGISTER = 0x04;
    public const UInt16 CM_RECEIVE = 0x05;
    public const UInt16 CM_EVENT = 0x06;
    public const UInt16 CM_NODUPLICATES = 0x10;
    public const UInt16 CM_SEEKPATTERN = 0x20;
    public const UInt16 CM_MISSPATTERN = 0x40;
    public const UInt16 CM_STREAM = 0x80;
    public const UInt16 CM_BCAST = 0x100;
    public const UInt16 CM_NETWORK = 0x200;
    public const UInt16 CM_MCAST = 0x200;
    public const UInt16 CM_USE_ON_ERROR = 0x400;
    public const UInt16 CM_USEONERROR = 0x400;
    public const UInt16 CM_RETRY = 0x800;
    public const UInt16 CM_GROUPED = 0x1000;
    public const UInt16 CM_CONNECT = 0x2000;
    public const UInt16 CM_WAIT = 0x4000;
    public const UInt16 CM_SYNCNOTIFY = 0x8000;
  }
  /// <summary>
  /// Available Data Array Types
  /// </summary>
  public static class ArrayType
  {
    public const UInt16 AT_UNKNOWN = 0x00;
    public const UInt16 AT_SCALAR = 0x01;
    public const UInt16 AT_SINGLE = 0x02;
    public const UInt16 AT_DOUBLE = 0x04;
    public const UInt16 AT_BOOLEAN = 0x08;
    public const UInt16 AT_CHANNEL = 0x10;
    public const UInt16 AT_SPECTRUM = 0x20;
    public const UInt16 AT_TRACE = 0x20;
    public const UInt16 AT_WAVEFORM = 0x20;
    public const UInt16 AT_STRUCTURED = 0x40;
    public const UInt16 AT_COLLECTION = 0x40;
    public const UInt16 AT_UNSTRUCTURED = 0x80;
    public const UInt16 AT_BLOB = 0x80;
    public const UInt16 AT_IMAGE = 0x80;
    public const UInt16 AT_JPEG = 0x100;
    public const UInt16 AT_MPEG = 0x200;
    public const UInt16 AT_BMP = 0x400;
    public const UInt16 AT_GIF = 0x800;
    public const UInt16 AT_WAV = 0x1000;
    public const UInt16 AT_MP3 = 0x2000;
    public static Boolean IsChannel(UInt16 code)
    {
      return ((code & AT_CHANNEL) == AT_CHANNEL);
    }
  }
  /// <summary>
  /// Available Access Flags
  /// </summary>
  public static class Access
  {
    public const UInt16 CA_NULL = 0x00;
    public const UInt16 CA_READ = 0x01;
    public const UInt16 CA_WRITE = 0x02;
    public const UInt16 CA_REPEAT = 0x04;
    public const UInt16 CA_SYNC = 0x08;
    public const UInt16 CA_ALARM = 0x10;
    public const UInt16 CA_HIST = 0x20;
    public const UInt16 CA_FIRST = 0x40;
    public const UInt16 CA_LAST = 0x80;
    public const UInt16 CA_RETRY = 0x100;
    public const UInt16 CA_NETWORK = 0x200;
    public const UInt16 CA_QUERY = 0x400;
    public const UInt16 CA_LOCKED = 0x800;// used only in server-side eqm dispatch
    public const UInt16 CA_STREAM = 0x800;// used only in client-side sync call
    public const UInt16 CA_XREAD = 0x800; // used only in property registration
    public const UInt16 CA_NORETRY = 0x1000;
    public const UInt16 CA_CONNECT = 0x2000;
    public const UInt16 CA_MUTABLE = 0x4000;
    public const UInt16 CA_SYNCNOTIFY = 0x8000;
    public static Boolean IsRead(UInt16 code)
    {
      return ((code & CA_READ) == CA_READ);
    }
    public static Boolean IsWrite(UInt16 code)
    {
      return ((code & CA_WRITE) == CA_WRITE);
    }
    public static Boolean IsExclusiveRead(UInt16 code)
    {
      return ((code & CA_XREAD) == CA_XREAD);
    }
    public static Boolean IsFirst(UInt16 code)
    {
      return ((code & CA_FIRST) == CA_FIRST);
    }
    public static Boolean IsLast(UInt16 code)
    {
      return ((code & CA_LAST) == CA_LAST);
    }
    public static Boolean IsLocked(UInt16 code)
    {
      return ((code & CA_LOCKED) == CA_LOCKED);
    }
  }
  /// <summary>
  /// Available Data Format Types
  /// </summary>
  public static class Formats
  {
    public const byte CF_DOUBLE = 0;
    public const byte CF_INT16  = 1;
    public const byte CF_BYTE = 2;
    public const byte CF_INT8 = 2;
    public const byte CF_INT32 = 3;
    public const byte CF_CHAR = 4;
    public const byte CF_FLOAT = 5;
    public const byte CF_INT64 = 6;
    public const byte CF_STRUCT = 7;
    public const byte CF_NAME8 = 8;
    public const byte CF_NAME16 = 9;
    public const byte CF_NAME16FI = 10;
    public const byte CF_BIT = 11;
    public const byte CF_NAME32 = 13;
    public const byte CF_FLTINT = 14;
    public const byte CF_INTINT = 15;
    public const byte CF_INTFLTINT = 18;
    public const byte CF_NAME48 = 19;
    public const byte CF_NAME8I = 20;
    public const byte CF_NAME16I = 21;
    public const byte CF_NAME32I = 22;
    public const byte CF_NAME48I = 23;
    public const byte CF_USTRING  = 24;
    public const byte CF_FLTFLTINT = 25;
    public const byte CF_FLTINTINT = 26;
    public const byte CF_FILTER = 27;
    public const byte CF_INTFLTFLTFLT = 27;
    public const byte CF_SPECTRUM = 28;
    public const byte CF_ADDRESS = 29;
    public const byte CF_INTINTINTINT = 29;
    public const byte CF_WINDOW = 30;
    public const byte CF_TTII = 30;
    public const byte CF_POINT = 31;
    public const byte CF_XY = 31;
    public const byte CF_FLTFLT = 31;
    public const byte CF_INTFLTSHTSHT = 32;
    public const byte CF_BOOLEAN = 33;
    public const byte CF_INTINTFLTFLT = 34;
    public const byte CF_FWINDOW = 34;
    public const byte CF_FLTINTFLTINT = 35;
    public const byte CF_NAME64 = 36;
    public const byte CF_NAME64I = 37;
    public const byte CF_DBLDBL = 40;
    public const byte CF_NAME16DBLDBL = 41;
    public const byte CF_NAME16INTINT = 42;
    public const byte CF_INTINTINT  = 43;
    public const byte CF_NAME32DBLDBL  = 44;
    public const byte CF_NAME64DBLDBL  = 45;
    public const byte CF_INTFLTFLTFLTNAME16 = 46;
    public const byte CF_UNAME = 46;
    public const byte CF_XML = 47;
    public const byte CF_DBLDBLDBL = 48;
    public const byte CF_NAME64DBLDBLDBL = 49;
    public const byte CF_BITFIELD8 = 50;
    public const byte CF_BITFIELD16 = 51;
    public const byte CF_BITFIELD32 = 52;
    public const byte CF_BITFIELD64 = 53;
    public const byte CF_NAME64DBL = 54;
    public const byte CF_IMAGE = 55;
    public const byte CF_HISTORY = 56;
    public const byte CF_STRING = 57;
    public const byte CF_ASPECTRUM = 58;
    public const byte CF_AIMAGE = 59;

    public const byte CF_DEFAULT = 254;
    public const byte CF_NULL = 255;

    public const short LFMT_OFFSET = 512;
    public static Int32 GetSizeInBytes(Int32 fmt)
    {
      return native.GetFormatSize(LFMT_OFFSET + (fmt%256));
    }
    public static Boolean IsPrimitive(Int32 fmt)
    {
      return (fmt <= CF_INT64);
    }
  }
  /// <summary>
  /// TINE Error definitions and flags
  /// </summary>
  public static class Errors
  {
    public const UInt16 CE_REDIRECTED  = 0x8000;
    public const UInt16 CE_SENDDATA    = 0x4000;
    public const UInt16 CE_RENEGOTIATE = 0x2000;

    public const UInt16 operation_success = 0; /**< successful completion */
    public const UInt16 illegal_line = 1;  /**< illegal line number */
    public const UInt16 illegal_format = 2;  /**< illegal format or data type */
    public const UInt16 illegal_arithmetic = 3;  /**< illegal arithmetic expression */
    public const UInt16 ambiguous = 4;  /**< ambigous input information */
    public const UInt16 illegal_delimiter = 5;  /**< illegal delimiter in expression */
    public const UInt16 zero_divide =  6;  /**< divide by zero */
    public const UInt16 work_area_full = 7;  /**< working area buffer is full */
    public const UInt16 non_existent = 8;  /**< requested target does not exist */
    public const UInt16 invalid_transport = 9;  /**< transport medium is not valid */
    public const UInt16 link_exhausted = 10; /**< link exhausted (unused) */

    public const UInt16 not_properly = 11; /**< expression not properly terminated (unused) */
    public const UInt16 un_allocated = 12; /**< requested target is not allocated in the given context */
    public const UInt16 no_such_line = 13; /**< expected line not found in database */
    public const UInt16 illegal_data_size = 14; /**< give data size is illegal or not valid */
    public const UInt16 io_error = 15; /**< bus or network io error */
    public const UInt16 illegal_context = 16; /**< given context is illegal */
    public const UInt16 runtime_error = 17; /**< runtime error */
    public const UInt16 system_error = 18; /**< system internal error */
    public const UInt16 hardware_busy = 19; /**< hardware is busy */
    public const UInt16 argument_list_error = 20; /**< input argument or parameter is invalid (alias for invalid_parameter */
    public const UInt16 invalid_parameter = 20; /**< input argument or parameter is invalid */

    public const UInt16 file_error =   21; /**< file access error */
    public const UInt16 use_stream_transport = 22; /**< datagram message size exceeds the allowable mtu (reserved) */
    public const UInt16 dimension_error = 23; /**< array dimension error */
    public const UInt16 square_root_negative = 24; /**< square root of a negative number */
    public const UInt16 buffer_too_small = 25; /**< supplied buffer is too small to handle the request */
    public const UInt16 string_too_long = 26; /**< supplied string is too long to buffer and cannot be truncated */
    public const UInt16 ipx_socket_error = 27; /**< ipx socket error (reserved) */
    public const UInt16 net_read_error = 28; /**< network read error */
    public const UInt16 not_ready = 29; /**< call to equipment module dispatch is not ready (internal, reserved) */
    public const UInt16 invalid_transport_size = 30; /**< total message size exceeds the configured allowable message size */

    public const UInt16 log_negative = 31; /**< log of a negative or zero number */
    public const UInt16 device_not_connected = 32; /**< requested device is not connected */
    public const UInt16 unauthorised_action = 33; /**< requested operation is not authorized */
    public const UInt16 hardware_error = 34; /**< hardware error */
    public const UInt16 illegal_equipment_number = 35; /**< requested device does not have a table entry (alias for illegal_device_number) */
    public const UInt16 illegal_device_number = 35; /**< requested device does not have a table entry */
    public const UInt16 illegal_device = 35; /**< requested device is illegal */
    public const UInt16 illegal_property = 36; /**< requested property is illegal */
    public const UInt16 out_of_range = 37; /**< given value is out of range */
    public const UInt16 not_implemented = 38; /**< requested operation has not been implemented */
    public const UInt16 no_such_computer = 39; /**< requested host computer does not exist */
    public const UInt16 struct_sealed = 40; /**< tagged structure has already been sealed (reserved) */

    public const UInt16 syntax_error = 41; /**< expression syntax error (unused) */
    public const UInt16 no_such_file = 42; /**< requested file does not exist */
    public const UInt16 file_already_exists = 43; /**< requested file already exists */
    public const UInt16 no_file_space = 44; /**< not enough capacity in pre-allocated random access file to complete the requested operation */
    public const UInt16 link_not_open = 45; /**< presistent link has timed out (alias for link_timeout) */
    public const UInt16 link_timeout = 45; /**< presistent link has timed out */
    public const UInt16 remitted_data_lost = 46; /**< returned only a portion of the requested data set (reserved) */
    public const UInt16 end_of_file =  47; /**< attempt to read past the end of a file (unused) */
    public const UInt16 archive_busy = 48; /**< archive operation is in progress */
    public const UInt16 server_name_in_use = 49; /**< device server name is already being used */
    public const UInt16 no_such_column = 50; /**< expected column not found in database */

    public const UInt16 out_of_client_memory = 51; /**< a client call could not allocate sufficient memory */
    public const UInt16 mcast_access_required = 52; /**< access requires CA_NETWORK which was not supplied! */
    public const UInt16 address_unresolved = 53; /**< requested target address could not be resolved */
    public const UInt16 invalid_property = 54; /**< requested property is not properly formed */
    public const UInt16 address_unknown = 55; /**< requested target address could not be found */
    public const UInt16 name_unknown = 56; /**< requested name could not be found */
    public const UInt16 invalid_keyword = 57; /**< input globals keyword is invalid */
    public const UInt16 invalid_link = 58; /**< input link id is invalid */
    public const UInt16 string_expected = 59; /**< database file read encountered an end-of-line while expecting a string */       
    public const UInt16 out_of_local_memory =    60; /**< a local call to allocated memory was not successful */

    public const UInt16 out_of_shared_memory = 61; /**< a call to obtain a shared-memory buffer was not successful */
    public const UInt16 invalid_structure_tag = 62; /**< the given structure tag is not in the structure registry */
    public const UInt16 invalid_index = 63; /**< the index given does not reference a table entry */
    public const UInt16 illegal_equipment_name = 64; /**< the equipment name given is not valid */
    public const UInt16 link_error = 65; /**< internal error processing data link (reserved) */
    public const UInt16 code_failure = 66; /**< internal failure due to unhandled exceptions, unavailable resource, OS failure, etc. */
    public const UInt16 non_existent_server = 67; /**< requested server does not exist */
    public const UInt16 deprecated = 68; /**< requested action or functionality has been deprecated */
    public const UInt16 not_supported = 69; /**< requested action or functionality not supported */
    public const UInt16 illegal_character =  70; /**< input character is illegal (unused) */

    public const UInt16 illegal_operator = 71; /**< input arithmetic or logical operator is not allowed */
    public const UInt16 not_allowed = 72; /**< requested operation is not allowed */
    public const UInt16 illegal_read_write = 73; /**< requested access is not allowed (typically CA_WRITE) */
    public const UInt16 out_of_server_memory = 74; /**< a server call could not allocate sufficient memory */
    public const UInt16 database_not_loaded = 75; /**< requested operation requires a database which has not been loaded */
    public const UInt16 illegal_command = 76; /**< requested command is illegal */
    public const UInt16 resources_exhausted = 77; /**< requested operation would exceed the configured capacity of an allocation table (clients, contracts, connections, etc.) */
    public const UInt16 file_not_open = 78; /**< requested file io could not be made as the file is not open (unused) */
    public const UInt16 sedac_error =  79; /**< sedac bus error */
    public const UInt16 no_such_loop = 80; /**< (unused) */

    public const UInt16 semaphore_error = 81; /**< a semaphore or mutex could not be obtained or released */
    public const UInt16 driver_not_installed = 82; /**< request requires a hardware driver which is not installed */
    public const UInt16 port_not_available = 83; /**< request requires access to a port whichis not available */
    public const UInt16 scan_error = 84; /**< (unused) */
    public const UInt16 semaphore_busy = 85; /**< a semahpre or mutex could not be obtained within the requested time */
    public const UInt16 non_existent_elem =  86; /**< the requested equipment modules does not exist in the registry */
    public const UInt16 non_existent_fec = 87; /**< the requested FEC does not exist */
    public const UInt16 non_existent_client = 88; /**< (unused) */
    public const UInt16 cannot_lock = 89; /**< not able to obtain an access lock or lock resources */
    public const UInt16 not_running = 90; /**< the requested resource is not running */

    public const UInt16 not_posted = 91; /**< a dispatch routine did not post a reply */
    public const UInt16 not_accepted = 92; /**< the requested operation was refused */
    public const UInt16 operation_timeout =  93; /**< the requested operation did not complete within the time limit specified */
    public const UInt16 illegal_protocol = 94; /**< the incoming tine protocol is not recognized (reserved, internal) */
    public const UInt16 gpib_error = 95; /**< gpib bus error */
    public const UInt16 rs232_error = 96; /**< rs232 bus error */
    public const UInt16 operation_busy = 97; /**< the requested operation is busy and needs more time to complete */
    public const UInt16 connection_timeout = 98; /**< a synchronous link did not return within the time limit specified */
    public const UInt16 illegal_mode = 99; /**< the requested access mode is not valid */
    public const UInt16 not_owner = 100; /**< (unused) */

    public const UInt16 not_defined = 101; /**< the requested operation has not yet been defined */
    public const UInt16 net_write_error = 102; /**< unable to send data on the network */
    public const UInt16 invalid_data = 103; /**< the input data are not valid */
    public const UInt16 software_error = 104; /**< general software error */
    public const UInt16 access_denied = 105; /**< call was refused by the security subsystem */
    public const UInt16 tcp_not_supported = 106; /**< an attempt to use TCP/IP access was made and is not supported on this host (reserved) */
    public const UInt16 ipx_not_supported	 = 107; /**< an attempt to use IPX access was made and is not supported on this host (reserved) */
    public const UInt16 host_not_resolved	 = 108; /**< the requested host could not be resolved */
    public const UInt16 tcp_connect_error	 = 109; /**< unable to connect to a TCP/IP socket (reserved) */
    public const UInt16 tcp_socket_error 	 = 110; /**< general TCP/IP socket error (reserved) */

    public const UInt16 configuration_error = 111; /**< resource configuration error (e.g. redirection information) */
    public const UInt16 invalid_return_code = 112; /**< indicates a systematic error in processing the equipment dispatch handler (reserved) */
    public const UInt16 link_blacklisted 	 = 113; /**< requested link parameters cannot succeed (due to initial return of non_existent_elem, address_unknown, etc.) */
    public const UInt16 link_exists = 114; /**< requested link already exists in the link table indicating a dependent link (reserved,internal) */
    public const UInt16 max_alarms_exceeded = 115; /**< attempt to add an alarm to the local alarm system whose capacity has been reached */
    public const UInt16 not_exported = 116; /**< (unused) */
    public const UInt16 is_switching = 117; /**< switching transition is in progress (unused) */
    public const UInt16 at_limit = 118; /**< setting is at a maximum or minimum (unused) */
    public const UInt16 get_subscription_id =   119; /**< used by access mode CM_NETWORK (reserved, internal) */
    public const UInt16 renegotiate_contract =  120; /**< used to repackage output data when few bytes are returned than requested (reserved, internal) */

    public const UInt16 server_redirection =    121; /**< used to supply redirection information (reserved, internal) */
    public const UInt16 value_too_high = 122; /**< returned value is too high */
    public const UInt16 value_too_low = 123; /**< returned value is too low  */
    public const UInt16 warn_too_high = 124; /**< returned value is approaching too high  */
    public const UInt16 warn_too_low = 125; /**< returned value is approaching too low  */
    public const UInt16 completion_waiting =    126; /**< dependent link is waiting for parent to complete */
    public const UInt16 cannot_select = 127; /**< select() operation failed (reserved) */
    public const UInt16 has_query_function =    128; /**< query call to PROPERTIES or DEVICES has completed but indicated a hierachical preference (reserved) */
    public const UInt16 not_signalled = 129; /**< contract is waiting to call the equipment module dispatch routine (reserved, internal) */
    public const UInt16 call_redirection =  130; /**< nominal alias for server_redirection */

    public const UInt16 udp_socket_error 	 =    131; /**< general UDP socket error (reserved) */
    public const UInt16 mutex_error = 132; /**< a mutex could not be obtained or released */
    public const UInt16 data_not_local = 133; /**< indicates that a device withing a wildcard call needs to span multiple servers (reserved, internal) */
    public const UInt16 name_already_exists =   134; /**< a structure or bitfield field name already exists */
    public const UInt16 already_assigned =  135; /**< the requested resource has already been assigned */
    public const UInt16 invalid_field = 136; /**< a structure or bitfield field is invalid */
    public const UInt16 mcast_init_error =  137; /**< could not initialize a multicast socket (reserved) */
    public const UInt16 invalid_mcast_group =   138; /**< multicast group address is invalid */
    public const UInt16 non_existent_function = 139; /**< an equipment module's dispatch routine is not registered */
    public const UInt16 property_is_mca = 140; /**< requested property is a multi-channel array (reserved, internal) */

    public const UInt16 invalid_name = 141; /**< input name is invalid */
    public const UInt16 invalid_value = 142; /**< input value is invalid */
    public const UInt16 device_error = 143; /**< requested device is in an error state */
    public const UInt16 device_not_ready = 144; /**< requested device is not ready */
    public const UInt16 device_busy = 145; /**< requested device is busy */
    public const UInt16 invalid_interval = 146; /**< requested polling interval is invalid */
    public const UInt16 notification_pending =  147; /**< a link is pending notification (reserved, internal) */
    public const UInt16 thread_error = 148; /**< a thread operation failed */
    public const UInt16 is_terminating = 149; /**< the device server is terminating (reserved, internal) */
    public const UInt16 name_server_unknown =   150; /**< the equipment name server address is unknown (reserved, internal) */

    public const UInt16 lock_required = 151; /**< an access lock is required to process the operation */
    public const UInt16 not_initialized = 152; /**< server is not yet initialized */
    public const UInt16 invalid_structure_size = 153; /**< the input structure size does not match the size in the structure registry */
    public const UInt16 canbus_error = 154; /**< can bus error */
    public const UInt16 profibus_error = 155; /**< profibus error */
    public const UInt16 data_stale =  156; /**< contract's data are stale (reserved) */
    public const UInt16 has_bitfield_tag = 157; /**< reqested contract needs to request a bitfield (reserved, internal) */
    public const UInt16 not_locked =  158; /**< the requested resource does not have an access lock (reserved) */
    public const UInt16 lock_expired = 159; /**< the requested access lock has expired (reserved) */

    public const UInt16 not_registered = 160; /**< the requested property (e.g. SystemScheduleProperty()) has not been registered */
    public const UInt16 non_existent_property = 161; /**< the requested property does not exist */
    public const UInt16 memory_overwrite = 162; /**< a heap error within a dispatch routine has been detected */
    public const UInt16 server_idle = 163; /**< the server is in an idle state */
    public const UInt16 not_applicable = 164; /**< the requested opertion is not applicable in the current context */
    public const UInt16 access_locked = 165; /**< access temporarily suspended due to access lock */
    public const UInt16 invalid_reference = 166; /**< invalid data reference handle */
    public const UInt16 has_structure_tag = 167; /**< reqested contract needs to request a structure (reserved, internal) */
    public const UInt16 warn_disk_space = 168; /**< warn that disk space is low (local alarm system) */
    public const UInt16 low_disk_space = 169; /**< signal that disk space is very low (local alarm system)  */

    public const UInt16 information_static = 170; /**< signal that static information is being 'polled'  */
    public const UInt16 reset_mca_property = 171; /**< signal to re-establish all mca single element links */
    public const UInt16 record_too_long = 172; /**< record length is too long to process within kernel */
    public const UInt16 unix_socket_error = 173; /**< an error condition on a unix (pipe) socket */
    public const UInt16 pipe_error = 173; /**< an error condition on a unix (pipe) socket */
    public const UInt16 shm_error = 174; /**< an error occured accessing shared memory */
    public const UInt16 tcp_connection_closed = 175; /**< tcp peer has closed the connection */
    public const UInt16 service_unavailable = 176; /**< requested service is unavailable */
    public const UInt16 device_offline = 177; /**< requested device is offline */
    public const UInt16 out_of_sequence = 178; /**< data value is out of sequence */
    public const UInt16 invalid_datarequest = 179; /**< requested data length or type is invalid (contract coercion) */

    private static sbyte[] errstr = new sbyte[64];
    public static String ToString(int err)
    {
      native.GetLastLinkError((short)err, errstr);
      char[] errchr = new char[64];
      for (int i = 0; i < 64 && errstr[i] != 0; i++) errchr[i] = (char)errstr[i];
      return new String(errchr);
    }
  }
  public static class StrUtils
  {
    public unsafe static string NameToString(byte* ba,int len)
    {
      char[] n = new char[len];
      int nc;
      for (nc = 0; nc < len && ba[nc] != 0; nc++) n[nc] = (char)ba[nc];
      return new String(n, 0, nc).Trim();
    }
    public unsafe static void StringToName(string s, byte* ba, int len)
    {
      int i = 0;
      if (s == null) return;
      foreach (char c in s)
      {
        ba[i++] = (byte)c;
        if (i >= len) break;
      }
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME8
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_NAME8;
    public fixed byte name[sizeInBytes];
    public NAME8(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_NAME16;
    public fixed byte name[sizeInBytes];
    public NAME16(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32
  {
    public const int sizeInBytes = 32;
    public const int format = Formats.CF_NAME32;
    public fixed byte name[sizeInBytes];
    public NAME32(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME48
  {
    public const int sizeInBytes = 48;
    public const int format = Formats.CF_NAME48;
    public fixed byte name[sizeInBytes];
    public NAME48(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64
  {
    public const int sizeInBytes = 64;
    public const int format = Formats.CF_NAME64;
    public fixed byte name[sizeInBytes];
    public NAME64(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_FLTINT;
    public float fval;
    public Int32 ival;
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_INTINT;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetInteger()
    {
      return i2val;
    }
    public void SetInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTFLT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_FLTFLT;
    public float f1val;
    public float f2val;
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DBLDBL
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_DBLDBL;
    public double d1val;
    public double d2val;
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME8I
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_NAME8I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16I
  {
    public const int sizeInBytes = 20;
    public const int format = Formats.CF_NAME16I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32I
  {
    public const int sizeInBytes = 36;
    public const int format = Formats.CF_NAME32I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME48I
  {
    public const int sizeInBytes = 52;
    public const int format = Formats.CF_NAME48I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64I
  {
    public const int sizeInBytes = 68;
    public const int format = Formats.CF_NAME64I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBL
  {
    public const int sizeInBytes = 72;
    public const int format = Formats.CF_NAME64DBL;
    public fixed byte name[sizeInBytes - 8];
    public double dval;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetDouble()
    {
      return dval;
    }
    public void SetDouble(double value)
    {
      dval = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16FI
  {
    public const int sizeInBytes = 24;
    public const int format = Formats.CF_NAME16FI;
    public fixed byte name[sizeInBytes - 8];
    public float fval;
    public Int32 ival;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16II
  {
    public const int sizeInBytes = 24;
    public const int format = Formats.CF_NAME16INTINT;
    public fixed byte name[sizeInBytes - 8];
    public Int32 i1val;
    public Int32 i2val;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public float GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16DBLDBL
  {
    public const int sizeInBytes = 32;
    public const int format = Formats.CF_NAME16DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32DBLDBL
  {
    public const int sizeInBytes = 48;
    public const int format = Formats.CF_NAME32DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBLDBL
  {
    public const int sizeInBytes = 80;
    public const int format = Formats.CF_NAME64DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTFLTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_INTFLTINT;
    public Int32 i1val;
    public float fval;
    public Int32 i2val;
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTFLTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_FLTFLTINT;
    public Int32 f1val;
    public float f2val;
    public Int32 ival;
    public Int32 GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(Int32 value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_FLTINTINT;
    public float fval;
    public Int32 i1val;
    public Int32 i2val;
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_INTINTINT;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 i3val;
    public Int32 GetTertiaryInteger()
    {
      return i3val;
    }
    public void SetTertiaryInteger(Int32 value)
    {
      i3val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DBLDBLDBL
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_DBLDBLDBL;
    public double d1val;
    public double d2val;
    public double d3val;
    public double GetTertiaryInteger()
    {
      return d3val;
    }
    public void SetTertiaryInteger(double value)
    {
      d3val = value;
    }
    public double GetPrimaryInteger()
    {
      return d1val;
    }
    public void SetPrimaryInteger(double value)
    {
      d1val = value;
    }
    public double GetSecondaryInteger()
    {
      return d2val;
    }
    public void SetSecondaryInteger(double value)
    {
      d2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTFLTFLTFLT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTFLTFLTFLT;
    public Int32 i1val;
    public float f1val;
    public float f2val;
    public float f3val;
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetTertiaryFloat()
    {
      return f3val;
    }
    public void SetTertiaryFloat(float value)
    {
      f3val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTINTINT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTINTINTINT;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 i3val;
    public Int32 i4val;
    public Int32 GetQuaternaryInteger()
    {
      return i4val;
    }
    public void SetQuaternaryInteger(Int32 value)
    {
      i4val = value;
    }
    public Int32 GetTertiaryInteger()
    {
      return i3val;
    }
    public void SetTertiaryInteger(Int32 value)
    {
      i3val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTFLTFLT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTINTFLTFLT;
    public Int32 i1val;
    public Int32 i2val;
    public float f1val;
    public float f2val;
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINTFLTINT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_FLTINTFLTINT;
    public float f1val;
    public Int32 i1val;
    public float f2val;
    public Int32 i2val;
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DWINDOW
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_WINDOW;
    public Int32 t1val;
    public Int32 t2val;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 GetPrimaryUTC()
    {
      return t1val;
    }
    public void SetPrimaryUTC(Int32 value)
    {
      t1val = value;
    }
    public Int32 GetSecondaryUTC()
    {
      return t2val;
    }
    public void SetSecondaryUTC(Int32 value)
    {
      t2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBLDBLDBL
  {
    public const int sizeInBytes = 88;
    public const int format = Formats.CF_NAME64DBLDBLDBL;
    public fixed byte name[sizeInBytes - 24];
    public double d1val;
    public double d2val;
    public double d3val;
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 24);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 24);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 24 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public double GetTertiaryDouble()
    {
      return d3val;
    }
    public void SetTertiaryDouble(double value)
    {
      d3val = value;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DUSTRING
  {
    public const int sizeInBytes = 96;
    public const int format = Formats.CF_USTRING;
    Int32 i1val;
    float f1val;
    float f2val;
    Int32 i2val;
    public fixed byte str[80];
    public string GetName()
    {
      fixed (byte* pca = str) return StrUtils.NameToString(pca, 80);
    }
    public void SetName(string str)
    {
      fixed (byte* pca = this.str) StrUtils.StringToName(str, pca, 80);
    }
    public double GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public double GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
  }

  //[StructLayout(LayoutKind.Explicit,Pack = 1,Size = 40+sizeof(IntPtr))]
  //public struct DTYPE
  //{
  //  [FieldOffset(0)]    int dArrayLength;
  //  [FieldOffset(4)]    short dFormat;
  //  [FieldOffset(6)]    short xferReason;
  //  [FieldOffset(8)]    int dStamp;
  //  [FieldOffset(12)]    int sysStamp;
  //  [FieldOffset(16)]    double dTimeStamp;
  //  [FieldOffset(24)]    char dTag;
  //  [FieldOffset(40)]    IntPtr data;
  //}
  //[StructLayout(LayoutKind.Sequential, Pack = 1)]
  //public unsafe struct DTYPE
  //{
  //  public int dArrayLength;
  //  public short dFormat;
  //  public short xferReason;
  //  public int dStamp;
  //  public int sysStamp;
  //  public double dTimeStamp;
  //  public fixed char dTag[Constants.TAG_NAME_SIZE];
  //  public IntPtr data;
  //}
  public static class AlarmDescriptors
  {
    public const UInt16 almNEWALARM = 0x01;
    public const UInt16 almHEARTBEAT = 0x02;
    public const UInt16 almOSCILLATION = 0x04;
    public const UInt16 almDATACHANGE = 0x08;
    public const UInt16 almTRANSIENT = 0x10;
    public const UInt16 almDISABLED = 0x20;
    public const UInt16 almTERMINATE = 0x40;
    public const UInt16 almTEST = 0x80;
    public const UInt16 almSUPPRESS = 0x80;
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DTYPE
  {
    public UInt32 dArrayLength;
    public short dFormat;
    public short xferReason;
    public int dStamp;
    public int sysStamp;
    public double dTimeStamp;
    //public fixed byte dTag[Constants.TAG_NAME_SIZE];
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] dTag;
    public IntPtr data;
  }

  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ALM_THRESHOLDS
  {
    public Int32 cnt;       /**< number successive threshold exceeded before applying alarm */
    public Int32 hisev;     /**< high severity */
    public Int32 hiwarnsev; /**< high warning severity */
    public Int32 losev;     /**< low severity */
    public Int32 lowarnsev; /**< low warning severity */
    public float hi;      /**< high threshold */
    public float hiwarn;  /**< high threshold warning (approaching high threshold) */
    public float lo;      /**< low threshold */
    public float lowarn;  /**< low threshold warning (approaching low threshold) */
    public UInt32 mask;   /**< if non-zero -> apply mask and check vs. normal value */
    public UInt32 normal; /**< normal valid data value */
    public Int32 normalIsAlarm; /**< apply a NOT to the 'normal value' */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ADS    /* Alarm Definition Structure */
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TAG_SIZE)]
      public char[] alarmTag;
    public UInt32 alarmCode;                /**< alarm code ID */
    public UInt32 alarmMask;                /**< alarm mask */
    public UInt16 alarmSystem;              /**< alarm system ID (0 -> use registered default) */
    public Int16 alarmSeverity;            /**< alarm severity (0 - 15) */
    public byte alarmDataFormat;            /**< alarm data format (simple code) */
    public byte alarmDataArraySize;         /**< alarm data array length */
    public byte alarmOscillationWindow;     /**< alarm oscillation window (in clear counts) */
    public byte alarmOscillationWindowPinned;/**< binary: 0 = not pinned */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] alarmText; /**< long text description for alarm tag */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] deviceText;/**< text description for device type */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] dataText;  /**< text description for alarm data contents */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_LONGSIZE)]
      public char[] url;   /**< url link to further information */
    public void initialize()
    {
      if (alarmTag == null) alarmTag = new char[Constants.ALARM_TAG_SIZE];
      if (alarmText == null) alarmText = new char[Constants.ALARM_TEXT_SIZE];
      if (deviceText == null) deviceText = new char[Constants.ALARM_TEXT_SIZE];
      if (dataText == null) dataText = new char[Constants.ALARM_TEXT_SIZE];
      if (url == null) url = new char[Constants.ALARM_TEXT_LONGSIZE];
    }
    public String GetAlarmTag()
    {
      return new String(alarmTag).Trim();
    }
    public void SetAlarmTag(String tag)
    {
      initialize();
      Array.Clear(alarmTag, 0, Constants.ALARM_TAG_SIZE);
      if (tag == null) tag = "";
      if (tag.Length > Constants.ALARM_TAG_SIZE) tag = tag.Substring(0, Constants.ALARM_TAG_SIZE);
      tag.ToCharArray().CopyTo(alarmTag, 0);
    }
    public String GetAlarmText()
    {
      return new String(alarmText).Trim();
    }
    public void SetAlarmText(String text)
    {
      initialize();
      Array.Clear(alarmText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(alarmText, 0);
    }
    public String GetDeviceText()
    {
      return new String(deviceText).Trim();
    }
    public void SetDeviceText(String text)
    {
      initialize();
      Array.Clear(deviceText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(deviceText, 0);
    }
    public String GetDataText()
    {
      return new String(dataText).Trim();
    }
    public void SetDataText(String text)
    {
      initialize();
      Array.Clear(dataText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(dataText, 0);
    }
    public String GetUrl()
    {
      return new String(url).Trim();
    }
    public void SetUrl(String url)
    {
      initialize();
      Array.Clear(this.url, 0, Constants.ALARM_TEXT_LONGSIZE);
      if (url == null) url = "";
      if (url.Length > Constants.ALARM_TEXT_LONGSIZE) url = url.Substring(0, Constants.ALARM_TEXT_LONGSIZE);
      url.ToCharArray().CopyTo(this.url, 0);
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct HistorySpecification
  {
    public Int32 pollingRate;      /**< polling rate in msec */
    public Int32 archiveRate;      /**< archive rate in sec */
    public Int32 depthShort;       /**< for short term storage */
    public Int32 depthLong;        /**< for long term storage */
    public Int32 heartbeat;        /**< archive heartbeat in sec */
    public float pTolerance;      /**< percent tolerance */
    public float aTolerance;      /**< absolute tolerance */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EXPORT_NAME_SIZE)]
    public char[] rhsServerName;   /**< Remote Server Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_NAME_SIZE)]
    public char[] rhsPropertyName; /**< Remote Property Name */
    public void initialize()
    {
      if (rhsServerName == null) rhsServerName = new char[Constants.EXPORT_NAME_SIZE];
      if (rhsPropertyName == null) rhsPropertyName = new char[Constants.PROPERTY_NAME_SIZE];
    }
    public void SetRemoteHistoryServerName(string ServerName)
    {
      initialize();
      if (ServerName == null) ServerName = "";
      if (ServerName.Length > Constants.EXPORT_NAME_SIZE) ServerName = ServerName.Substring(0, Constants.EXPORT_NAME_SIZE);
      ServerName.ToCharArray().CopyTo(rhsServerName, 0);
    }
    public void SetRemoteHistoryPropertyName(string PropertyName)
    {
      initialize();
      if (PropertyName == null) PropertyName = "";
      if (PropertyName.Length > Constants.PROPERTY_NAME_SIZE) PropertyName = PropertyName.Substring(0, Constants.PROPERTY_NAME_SIZE);
      PropertyName.ToCharArray().CopyTo(rhsPropertyName, 0);
    }
  } 
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct sockaddr_in
  {
    public Int16 sin_family;
    public UInt16 sin_port;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
      public byte[] sin_addr;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
     public char[] sin_zero;
    public void initialize()
    {
      if (sin_addr == null) sin_addr = new byte[4];
      if (sin_zero == null) sin_zero = new char[8];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ClnInfo
  {
    public UInt16 accessMode; /**< this client's requested access mode: CM_SINGLE, CM_DATACHANGE, CM_TIMER, etc. */
    public UInt16 reserved1;
    public Int32 pollingRate;/**< this client's requested polling rate */
    public UInt32 starttime;  /**< this client's start of initial subscription */
    public UInt16 counter;    /**< this client's counts remaining in subscription(s) */
    public UInt16 reserved2;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.USERNAME_SIZE)]
      public char[] userName; /**< this client's user name */
    public sockaddr_in addr; /**< this client's ip address */
    public short ncontracts;  /**< this client's total number of contracts */
    public short tineProtocol;/**< this client's tine protocol level */
    public short inetProtocol;/**< this client's transport protocol: UDP, TCP, STREAM , etc.  */
    public UInt16 reserved3;
    public void initialize()
    {
      if (userName == null) userName = new char[Constants.USERNAME_SIZE];
      addr.initialize();
    }
    public String GetUserName()
    {
      if (userName == null) return null;
      int len = 0;
      while (len < Constants.USERNAME_SIZE && userName[len] != 0) len++;
      return new String(userName,0,len);
    }
    public String GetNetworkAddr()
    {
      if (addr.sin_addr == null) return null;
      ushort p = addr.sin_port;
      p = (ushort)(((p & 0xffff) << 8) + (p >> 8));
      String ip = "" + addr.sin_addr[0] + "." + addr.sin_addr[1] + "." + addr.sin_addr[2] + "." + addr.sin_addr[3] + ":"+p;
      return ip;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct PrpQueryStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_NAME_SIZE)]
      public char[] prpName; /**< the property name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_DESC_SIZE)]
      public char[] prpDescription; /**< the property description */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_REDIR_SIZE)]
      public char[] prpRedirection; /**< a redirection string if the property is redirected to another server */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] prpTag;   /**< the data tag for output data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] prpTagIn; /**< the data tag for input data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.UNITS_SIZE)]
      public char[] prpUnits;   /**< a string containing the natural units of the data returned by the property */
    public float prpMinValue;    /**< the lower limit of valid data associated with the property */
    public float prpMaxValue;    /**< the upper limit of valid data associated with the property */
    public UInt32 prpSize;       /**< the maximum allowed size of an output data set associated with the property */
    public UInt32 prpSizeIn;     /**< the maximum allowed size of an input data set associated with the property */
    public UInt32 prpNumOverloads; /**< the number of overloads associated with this property name  */
    public UInt16 prpHistoryDepthShort; /**< the depth of the short term history kept for this property */
    public UInt16 prpHistoryDepthLong; /**< the depth of the long term history kept for this property */
    public byte prpFormat;       /**< the default output data format accepted for this property */
    public byte prpFormatIn;     /**< is the default input data format accepted for this property */
    public byte prpAccess;       /**< the allowed data access for this property (either CA_READ, CA_WRITE or both) */
    public byte prpGraphType;    /**< is the preferred graph type for output data associated with this property, e.g. GT_LIN, GT_LOG, etc */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.UNITS_SIZE)]
       public char[] rngUnits;   /**< a string containing the natural units of the x-axis (range) in case of a spectrum type array */
    public float rngMinValue;    /**< the lower limit of x-axis range associated with the property  */
    public float rngMaxValue;    /**< the upper limit of x-axis range associated with the property  */
    public UInt16 numRows;       /**< the number of rows in an array property */
    public UInt16 rowSize;       /**< the size of a row in an array property */
    public UInt16 prpArrayType;  /**< the property array type (AT_NONE, AT_SCALAR, AT_CHANNEL, AT_SPECTRUM, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
      public UInt16[] reserved;   /**< reserved field) */
    public static string getStructTag()
    {
      return "PRPQSr4";
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ExpDataStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] FecName;         /**< FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.SUBSYSTEM_NAME_SIZE)]
    public char[] SubSystem; /**< sub-system name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EXPORT_NAME_SIZE)]
    public char[] ExportName;   /**< Export Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EQM_NAME_SIZE)]
    public char[] EqmName;         /**< Equipment name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.CONTEXT_NAME_SIZE)]
    public char[] EqmContext;  /**< EqmContext */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FecDataStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] FecName;      /**< FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
    public byte[] IPXNetwork;  /**< IPX addr info (or internal) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
    public byte[] IPXNode;     /**< IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
    public byte[] IPXImmediateAddress; /**< IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ADDR_SIZE)]
    public char[] IP;          /**< IP address as string */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
    public byte[] IPh_addr;    /**< IP haddr (byte representation) */
    public Int32 portOffset;   /**< port offset (applied to all listening ports) */
    public Int32 inetProtocol; /**< inet protocol (UDP, TCP, IPX, MMF, PIPE, etc.) */
    public Int32 tineProtocol; /**< tine protocol level (highest level accepted) */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FecAddrStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] fecName;      /**< FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
    public byte[] netAdr;  /**< IPX addr info (or internal) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
    public byte[] lclAdr;     /**< IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 40)]
    public char[] strAdr;          /**< IP address as string */
    public Int32 portOffset;   /**< port offset (applied to all listening ports) */
    public Int32 inetProtocol; /**< inet protocol (UDP, TCP, IPX, MMF, PIPE, etc.) */
    public Int32 tineProtocol; /**< tine protocol level (highest level accepted) */
    public Int32 reserved; /**< tine protocol level (highest level accepted) */
  }
  class MemMgr
  {
    public object mmRef; // managed memory reference to object
    public unsafe IntPtr umRef; // unmanaged pointer reference;
    public MemMgr(object objRef, int dataSize) 
    {
      mmRef = objRef;
      umRef = Marshal.AllocHGlobal(dataSize);
    }
    public bool Equals(MemMgr mm) { return mm == null ? false : mm.mmRef == this.mmRef; }
  }
  /// <summary>
  /// TINE Data Type object: All data transfered client <-> server must be contained in a TDataType object.
  /// </summary>
  public class TDataType
  {
    private static List<MemMgr> memLst = new List<MemMgr>();
    private static IntPtr GetIntPtrFromManagedDataObject(object data,int dataSize)
    {
      MemMgr mm = memLst.Find(delegate(MemMgr m) { return m.mmRef == data; });
      if (mm == null)
      {
        mm = new MemMgr(data, dataSize);
        memLst.Add(mm);
      }
      return mm.umRef;
    }
    /// <summary>
    /// gets the TINE data format code for the object passed
    /// </summary>
    /// <param name="data">is the data object whose TINE data format is desired.</param>
    /// <returns>the TINE data format code assigned to the object specified</returns>
    public static byte GetTineFormat(Object data)
    {
      return GetTineType(data);
    }
    /// <summary>
    /// gets the TINE data format code for the Type passed
    /// </summary>
    /// <param name="type">is the .NET Type whose TINE data format is desired.</param>
    /// <returns>the TINE data format code assigned to the object specified</returns>
    public static byte GetTineFormat(Type type)
    {
      switch (Type.GetTypeCode(type))
      {
        case TypeCode.Boolean: return Formats.CF_BOOLEAN;
        case TypeCode.Byte: return Formats.CF_BYTE;
        case TypeCode.Char: return Formats.CF_CHAR;
        case TypeCode.Double: return Formats.CF_DOUBLE;
        case TypeCode.Empty: return Formats.CF_NULL;
        case TypeCode.Int16: return Formats.CF_INT16;
        case TypeCode.Int32: return Formats.CF_INT32;
        case TypeCode.Int64: return Formats.CF_INT64;
        case TypeCode.Single: return Formats.CF_FLOAT;
        case TypeCode.String: return Formats.CF_STRING;
        case TypeCode.UInt16: return Formats.CF_INT16;
        case TypeCode.UInt32: return Formats.CF_INT32;
        case TypeCode.UInt64: return Formats.CF_INT64;
        default: break;
      }
      return Formats.CF_NULL;
    }
    private static bool isSystemStruct(string cname)
    {
      if (cname.CompareTo("PrpQueryStruct") == 0) return true;
      return false;
    }
    private static string getSystemStructName(string cname)
    {
      if (native.GetStructSize(cname) > 0) return cname;
      // check system struct registry here :
      if (cname.CompareTo("PrpQueryStruct") == 0) return PrpQueryStruct.getStructTag();
      return null;
    }
    private static string getStructName(Object data)
    {
      if (data == (Object)null) return null;
      string cname = (data is TTaggedStruct) ? ((TTaggedStruct)data).GetTag() :
                     (data is TTaggedStruct[]) ? ((TTaggedStruct[])data)[0].GetTag() : data.GetType().Name;
      if (data is Array) cname = ((Array)data).GetType().GetElementType().Name;
      return getSystemStructName(cname);
    }
    private static byte GetTineType(Object data)
    {
      if (data == (Object)null) return Formats.CF_NULL;
      if (getStructName(data) != null) return Formats.CF_STRUCT;

      Type dtype;
      if (data is Array)
      {
        dtype = data.GetType().GetElementType();
      }
      else
      {
        dtype = data.GetType();
      }
      switch (Type.GetTypeCode(dtype))
      {
        case TypeCode.Boolean: return Formats.CF_BOOLEAN;
        case TypeCode.Byte: return Formats.CF_BYTE;
        case TypeCode.Char: return Formats.CF_CHAR;
        case TypeCode.Double: return Formats.CF_DOUBLE;
        case TypeCode.Empty: return Formats.CF_NULL;
        case TypeCode.Int16: return Formats.CF_INT16;
        case TypeCode.Int32: return Formats.CF_INT32;
        case TypeCode.Int64: return Formats.CF_INT64;
        case TypeCode.Single: return Formats.CF_FLOAT;
        case TypeCode.String: return Formats.CF_STRING;
        case TypeCode.UInt16: return Formats.CF_INT16;
        case TypeCode.UInt32: return Formats.CF_INT32;
        case TypeCode.UInt64: return Formats.CF_INT64;
        case TypeCode.Object:
          if (data is IntPtr) return Formats.CF_NULL;
          if (data is NAME8[] || data is NAME8) return Formats.CF_NAME8;
          if (data is NAME16[] || data is NAME16) return Formats.CF_NAME16;
          if (data is NAME32[] || data is NAME32) return Formats.CF_NAME32;
          if (data is NAME48[] || data is NAME48) return Formats.CF_NAME48;
          if (data is NAME64[] || data is NAME64) return Formats.CF_NAME64;
          if (data is NAME8I[] || data is NAME8I) return Formats.CF_NAME8I;
          if (data is NAME16[] || data is NAME16) return Formats.CF_NAME16;
          if (data is NAME32[] || data is NAME32) return Formats.CF_NAME32;
          if (data is NAME48[] || data is NAME48) return Formats.CF_NAME48;
          if (data is NAME64[] || data is NAME64) return Formats.CF_NAME64;
          if (data is FLTINT[] || data is FLTINT) return Formats.CF_FLTINT;
          if (data is INTINT[] || data is INTINT) return Formats.CF_INTINT;
          if (data is FLTFLT[] || data is FLTFLT) return Formats.CF_FLTFLT;
          if (data is DBLDBL[] || data is DBLDBL) return Formats.CF_DBLDBL;
          if (data is NAME8I[] || data is NAME8I) return Formats.CF_NAME8I;
          if (data is NAME16I[] || data is NAME16I) return Formats.CF_NAME16I;
          if (data is NAME32I[] || data is NAME32I) return Formats.CF_NAME32I;
          if (data is NAME48I[] || data is NAME48I) return Formats.CF_NAME48I;
          if (data is NAME64I[] || data is NAME64I) return Formats.CF_NAME64I;
          if (data is NAME64DBL[] || data is NAME64DBL) return Formats.CF_NAME64DBL;
          if (data is NAME16FI[] || data is NAME16FI) return Formats.CF_NAME16FI;
          if (data is NAME16II[] || data is NAME16II) return Formats.CF_NAME16INTINT;
          if (data is NAME16DBLDBL[] || data is NAME16DBLDBL) return Formats.CF_NAME16DBLDBL;
          if (data is NAME32DBLDBL[] || data is NAME32DBLDBL) return Formats.CF_NAME32DBLDBL;
          if (data is NAME64DBLDBL[] || data is NAME64DBLDBL) return Formats.CF_NAME64DBLDBL;
          if (data is INTFLTINT[] || data is INTFLTINT) return Formats.CF_INTFLTINT;
          if (data is FLTFLTINT[] || data is FLTFLTINT) return Formats.CF_FLTFLTINT;
          if (data is FLTINTINT[] || data is FLTINTINT) return Formats.CF_FLTINTINT;
          if (data is INTINTINT[] || data is INTINTINT) return Formats.CF_INTINTINT;
          if (data is DBLDBLDBL[] || data is DBLDBLDBL) return Formats.CF_DBLDBLDBL;
          if (data is INTINTINTINT[] || data is INTINTINTINT) return Formats.CF_INTINTINTINT;
          if (data is INTINTFLTFLT[] || data is INTINTFLTFLT) return Formats.CF_INTINTFLTFLT;
          if (data is INTFLTFLTFLT[] || data is INTFLTFLTFLT) return Formats.CF_INTFLTFLTFLT;
          if (data is FLTINTFLTINT[] || data is FLTINTFLTINT) return Formats.CF_FLTINTFLTINT;
          if (data is DWINDOW[] || data is DWINDOW) return Formats.CF_WINDOW;
          if (data is NAME64DBLDBLDBL[] || data is NAME64DBLDBLDBL) return Formats.CF_NAME64DBLDBLDBL;
          if (data is DUSTRING[] || data is DUSTRING) return Formats.CF_USTRING;
          if (data is TTaggedStruct[] || data is TTaggedStruct) return Formats.CF_STRUCT;
          break;
        default: break;
      }
      return Formats.CF_NULL;
    }
    public static unsafe IntPtr GetDataPtr(Object data)
    {
      return GetTineDataPtr(data, null);
    }
    public static unsafe IntPtr GetDataPtr(Object data,string tag)
    {
      return GetTineDataPtr(data,tag);
    }
    private static unsafe IntPtr GetTineDataPtr(Object data)
    {
      return GetTineDataPtr(data, null);
    }
    private static unsafe IntPtr GetTineDataPtr(Object data, string cname)
    {
      if (data == null) return IntPtr.Zero;
      Boolean isArray = data is Array;
      if (isArray && ((Array)data).Rank != 1)
      {
        throw new Exception("object is NOT a 1 dimensional array !");
      }
      Type dtype = data.GetType().GetElementType();
      switch (Type.GetTypeCode(dtype))
      {
        //case TypeCode.Boolean:
        //  fixed (double* p = (double[])data) return new IntPtr((void*)p);
        case TypeCode.Byte:
          fixed (byte* p = (byte[])data) return new IntPtr((void*)p);
        case TypeCode.Char:
          fixed (char* p = (char[])data) return new IntPtr((void*)p);
        case TypeCode.Double:
          fixed (double *p = (double[])data) return new IntPtr((void *)p);
        case TypeCode.Empty:
          if (data is TTaggedStruct)
          {
            cname = ((TTaggedStruct)data).GetTag();
            goto case TypeCode.Object;
          }
          if (data != null && !isArray)
          {
            throw new Exception("data object argument is NOT an array and is not passed by reference !");
          }
          return IntPtr.Zero;
        case TypeCode.Int16:
          fixed (Int16* p = (Int16[])data) return new IntPtr((void*)p);
        case TypeCode.Int32:
          fixed (Int32* p = (Int32[])data) return new IntPtr((void*)p);
        case TypeCode.Int64:
          fixed (Int64* p = (Int64[])data) return new IntPtr((void*)p);
        case TypeCode.Single:
          fixed (float* p = (float[])data) return new IntPtr((void*)p);
        //case TypeCode.String:
        //  fixed (double* p = (double[])data) return new IntPtr((void*)p);
        case TypeCode.UInt16:
          fixed (UInt16* p = (UInt16[])data) return new IntPtr((void*)p);
        case TypeCode.UInt32:
          fixed (UInt32* p = (UInt32[])data) return new IntPtr((void*)p);
        case TypeCode.UInt64:
          fixed (UInt64* p = (UInt64[])data) return new IntPtr((void*)p);
        case TypeCode.Object:
          if (data is NAME8[]) fixed (NAME8* p = &(((NAME8[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16[]) fixed (NAME16* p = &(((NAME16[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32[]) fixed (NAME32* p = &(((NAME32[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME48[]) fixed (NAME48* p = &(((NAME48[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64[]) fixed (NAME64* p = &(((NAME64[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINT[]) fixed (FLTINT* p = &(((FLTINT[])data)[0])) { return new IntPtr((void *)p); }
          if (data is INTINT[]) fixed (INTINT* p = &(((INTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTFLT[]) fixed (FLTFLT* p = &(((FLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DBLDBL[]) fixed (DBLDBL* p = &(((DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME8I[]) fixed (NAME8I* p = &(((NAME8I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16I[]) fixed (NAME16I* p = &(((NAME16I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32I[]) fixed (NAME32I* p = &(((NAME32I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME48I[]) fixed (NAME48I* p = &(((NAME48I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64I[]) fixed (NAME64I* p = &(((NAME64I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBL[]) fixed (NAME64DBL* p = &(((NAME64DBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16FI[]) fixed (NAME16FI* p = &(((NAME16FI[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16II[]) fixed (NAME16II* p = &(((NAME16II[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16DBLDBL[]) fixed (NAME16DBLDBL* p = &(((NAME16DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32DBLDBL[]) fixed (NAME32DBLDBL* p = &(((NAME32DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBLDBL[]) fixed (NAME64DBLDBL* p = &(((NAME64DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTFLTINT[]) fixed (INTFLTINT* p = &(((INTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTFLTINT[]) fixed (FLTFLTINT* p = &(((FLTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINTINT[]) fixed (FLTINTINT* p = &(((FLTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTINT[]) fixed (INTINTINT* p = &(((INTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DBLDBLDBL[]) fixed (DBLDBLDBL* p = &(((DBLDBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTINTINT[]) fixed (INTINTINTINT* p = &(((INTINTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTFLTFLTFLT[]) fixed (INTFLTFLTFLT* p = &(((INTFLTFLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTFLTFLT[]) fixed (INTINTFLTFLT* p = &(((INTINTFLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINTFLTINT[]) fixed (FLTINTFLTINT* p = &(((FLTINTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DWINDOW[]) fixed (DWINDOW* p = &(((DWINDOW[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBLDBLDBL[]) fixed (NAME64DBLDBLDBL* p = &(((NAME64DBLDBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DUSTRING[]) fixed (DUSTRING* p = &(((DUSTRING[])data)[0])) { return new IntPtr((void*)p); }
          //if (data is TTaggedStruct[])
          if (cname != null || dtype.IsValueType)
          {
            if (cname == null) cname = dtype.IsArray ? dtype.GetElementType().Name : dtype.Name;
            int ssiz;
            if ((ssiz = native.GetStructSize(cname)) > 0)
            {
              if (isArray)
              {
                return GetIntPtrFromManagedDataObject(data, ssiz * ((Array)data).Length);
              }
              else
              {
                return GetIntPtrFromManagedDataObject(data, ssiz);
              }
            }
          }
          break;
        //  default: return Formats.CF_NULL;
      }
      return IntPtr.Zero;
    }
    private UInt32 GetTineArrayLength(Object data)
    {
      if (data == null) return 0;
      if (data is System.Array)
      {
        Array da = (Array)data;
        //if (da.Rank != 1) return 0;
        //return (UInt32)da.GetLength(0);
        int dim = da.Rank;
        int len = 1;
        while (--dim >= 0) len *= da.GetLength(dim);
        return (UInt32)len;
      }
      return 1;
    }
    private unsafe DTYPE tdts = new DTYPE();
    private Object data;
    public Boolean isByRef = false;
    public Object GetDataObject() { return data; }
    public unsafe DTYPE GetDTYPE() 
    {
      return tdts; 
    }
    private unsafe IntPtr GetTineDataPtr(ref double data)
    {
      fixed (double* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref float data)
    {
      fixed (float* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref byte data)
    {
      fixed (byte* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int16 data)
    {
      fixed (Int16* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt16 data)
    {
      fixed (UInt16* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int32 data)
    {
      fixed (Int32* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt32 data)
    {
      fixed (UInt32* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int64 data)
    {
      fixed (Int64* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt64 data)
    {
      fixed (UInt64* p = &data) return new IntPtr((void*)p);
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref double data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_DOUBLE);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref float data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned); //n.b. this only pins a new object boxing the value
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_FLOAT);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt16 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT16);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int16 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT16);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt32 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT32);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int32 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT32);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt64 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT64);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int64 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT64);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    private void AssignCompoundDataType(object data,byte fmt,int siz)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = 1;
      tdts.data = GetIntPtrFromManagedDataObject(data, siz);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    public unsafe TDataType(ref NAME64 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64, NAME64.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32, NAME32.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16, NAME16.sizeInBytes);
    }
    public unsafe TDataType(ref NAME8 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME8, NAME8.sizeInBytes);
    }
    public unsafe TDataType(ref FLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTINT, FLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref INTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_INTINT, INTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTFLT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTFLT, FLTFLT.sizeInBytes);
    }
    public unsafe TDataType(ref DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_DBLDBL, DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64I, NAME64I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64DBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64DBL, NAME64DBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32I, NAME32I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16I, NAME16I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16FI data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16FI, NAME16FI.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64DBLDBL, NAME64DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32DBLDBL, NAME32DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref INTFLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_INTFLTINT, INTFLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTFLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTFLTINT, FLTFLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTINTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTINTINT, FLTINTINT.sizeInBytes);
    }
    public unsafe TDataType(ref DUSTRING data)
    {
      AssignCompoundDataType(data, Formats.CF_USTRING, DUSTRING.sizeInBytes);
    }
    private GCHandle gchData;
    private Object gchRef = null;
    public Object GetReferencedDataObject()
    {
      return gchRef;
    }
    private void SetAsNullType()
    {
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_NULL);
      tdts.dArrayLength = 0;
      tdts.data = IntPtr.Zero;
    }
    private static TDataType dtNull = new TDataType(null);
    public static TDataType empty()
    {
      return dtNull;
    }
    /// <summary>
    /// makes a TDataType from the .NET object passed
    /// </summary>
    /// <param name="data">the .NET object which should be bound to the TDataType</param>
    public TDataType(Object data)
    {
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      //if (!(data is Array)) throw exception
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);
      if (data is Array)
      { // whatever the rank, do it this way ...
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
        tdts.data = gchData.AddrOfPinnedObject();
      }
      else
      {
        tdts.data = GetTineDataPtr(data);
      }
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i=0; i<ch.Length && i<Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      else if (gchData == null || !gchData.IsAllocated)
      {
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      }
      this.data = data;
    }
    public TDataType(Array data)
    {
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);

      if (fmt == Formats.CF_STRUCT)
      {
        tdts.data = GetTineDataPtr(data);
      }
      else
      {
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
        tdts.data = gchData.AddrOfPinnedObject();
      }
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i = 0; i < ch.Length && i < Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      this.data = data;
    }
    private TDataType(Array data,int row,int offset,int length)
    { // this probably won't work ...
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      if (!Formats.IsPrimitive(fmt))
      {
        if (row != 0 || offset != 0)
        {
          throw new Exception("Cannot parse a non-primitive data type !");
        }
      }
      if (data.Rank > 2)
      {
        throw new Exception("arrays with more than 2 dimensions not supported!");
      }
      int rowlen = data.GetLength(row);
      if (row >= data.Rank) row = data.Rank - 1;
      if (length > rowlen) length = rowlen;
      if (offset >= length)
      {
        throw new Exception("offset is larger than the row length !");
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      tdts.data = gchData.AddrOfPinnedObject();
      int p = row * rowlen + offset;
//      tdts.data = tdts.data + (IntPtr)p;

      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i = 0; i < ch.Length && i < Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      this.data = data;
    }
    public TDataType(DTYPE dt)
    {
      tdts.dFormat = dt.dFormat;
      tdts.dArrayLength = dt.dArrayLength;
      tdts.dStamp = dt.dStamp;
      tdts.dTimeStamp = dt.dTimeStamp;
      tdts.sysStamp = dt.sysStamp;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dt.dTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dt.dTag.Length < siz) siz = dt.dTag.Length;
        Array.Copy(dt.dTag, tdts.dTag, siz);
      }
      tdts.data = dt.data; // the underlying pointer
      this.data = dt.data; // the app's pointer
    }
    public TDataType(DTYPE dt,object dobj)
    {
      tdts.dFormat = dt.dFormat;
      tdts.dArrayLength = dt.dArrayLength;
      tdts.dStamp = dt.dStamp;
      tdts.dTimeStamp = dt.dTimeStamp;
      tdts.sysStamp = dt.sysStamp;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dt.dTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dt.dTag.Length < siz) siz = dt.dTag.Length;
        Array.Copy(dt.dTag, tdts.dTag, siz);
        gchRef = dobj;
      }
      tdts.data = dt.data;
      this.data = dt.data;
    }
    public TDataType(int dataSize, byte dataFmt, string dataTag)
    {
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + dataFmt);
      tdts.dArrayLength = (uint)dataSize;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dataTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dataTag.Length < siz) siz = dataTag.Length;
        Array.Copy(dataTag.ToCharArray(), tdts.dTag, siz);
      }
    //  tdts.data = IntPtr.Zero;
    //  this.data = IntPtr.Zero;
      data = new byte[dataSize * Formats.GetSizeInBytes(dataFmt)];
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      tdts.data = GetTineDataPtr(data);
    }
    ~TDataType()
    {
      if (gchData != null && gchData.IsAllocated) gchData.Free();
    }
    public Int32 GetData(ref Int16 data) 
    {
      Int16[] v = new Int16[1];
      Int32 cc = native.GetValuesAsShort(ref tdts,v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref UInt16 data) 
    {
      Int16[] sdata = new Int16[1];
      Int32 cc = native.GetValuesAsShort(ref tdts,sdata, 1);
      data = (UInt16)sdata[0];
      return cc;
    }
    public Int32 GetData(ref Int32 data) 
    { 
      Int32[] v = new Int32[1];
      Int32 cc = native.GetValuesAsLong(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref UInt32 data) 
    {
      Int32[] sdata = new Int32[1];
      Int32 cc = native.GetValuesAsLong(ref tdts, sdata, 1);
      data = (UInt32)sdata[0];
      return cc;
    }
    public Int32 GetData(ref float data) 
    {
      float[] v = new float[1];
      Int32 cc = native.GetValuesAsFloat(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref double data) 
    {
      double[] v = new double[1];
      Int32 cc = native.GetValuesAsDouble(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(Object data)
    {
      if (data is Array) return GetData(data, ((Array)data).Length, 0);
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);
      Type dtype = data.GetType();
      if (!dtype.IsByRef) return Errors.invalid_parameter;

      switch (fmt)
      {
        case Formats.CF_STRUCT:
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            data = Marshal.PtrToStructure(tdts.data, dtype);
          }
          cc = 0;
          break;
        default:
          {
            int siz = Formats.GetSizeInBytes(fmt);
            IntPtr value = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, value, (short)(Formats.LFMT_OFFSET + fmt), siz, 1, 0);
            break;
          }
      }
      return cc;
    }
    public Int32 GetData(Object data,int length,int offset)
    {
      if (length <= 0) return Errors.dimension_error;
      if (!(data is Array)) return Errors.dimension_error;
      if (((Array)data).Rank != 1) return Errors.dimension_error;
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);

      switch (fmt)
      {
        case Formats.CF_BYTE:
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (offset == 0) return native.GetValuesAsByte(ref tdts, (byte[])data, length);
            byte[] values = new byte[length];
            cc = native.GetValuesAsByte(ref tdts, values, length);
            System.Array.Copy(values, 0, (byte[])data, offset, length);
          }
          break;
        case Formats.CF_CHAR:
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (offset == 0) return native.GetValuesAsStringEx(ref tdts, (char[])data, Formats.CF_CHAR, length, 0);
            char[] values = new char[length];
            cc = native.GetValuesAsStringEx(ref tdts, values, Formats.CF_CHAR, length, 0);
            System.Array.Copy(values, 0, (char[])data, offset, length);
          }
          break;
        case Formats.CF_INT16:
          {
            if (offset == 0) return native.GetValuesAsShort(ref tdts, (Int16[])data, length);
            Int16[] values = new Int16[length];
            cc = native.GetValuesAsShort(ref tdts, values, length);
            System.Array.Copy(values, 0, (Int16[])data, offset, length);
          }
          break;
        case Formats.CF_INT32:
          {
            if (offset == 0) return native.GetValuesAsLong(ref tdts, (Int32[])data, length);
            Int32[] values = new Int32[length];
            cc = native.GetValuesAsLong(ref tdts, values, length);
            System.Array.Copy(values, 0, (Int32[])data, offset, length);
          }
          break;
        case Formats.CF_INT64:
          {
            IntPtr values = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 8, length, offset);
          }
          break;
        case Formats.CF_FLOAT:
          {
            if (offset == 0) return native.GetValuesAsFloat(ref tdts, (float[])data, length);
            float[] values = new float[length];
            cc = native.GetValuesAsFloat(ref tdts, values, length);
            System.Array.Copy(values, 0, (float[])data, offset, length);
          }
          break;
        case Formats.CF_DOUBLE:
          {
            if (offset == 0) return native.GetValuesAsDouble(ref tdts, (double[])data, length);
            double[] values = new double[length];
            cc = native.GetValuesAsDouble(ref tdts, values, length);
            System.Array.Copy(values, 0, (double[])data, offset, length);
          }
          break;
        case Formats.CF_STRUCT:
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            int n = 0, alen = ((Array)data).Length;
            object obj;
            IntPtr p = tdts.data;
            for (int i = 0; i < alen; i++)
            {
              obj = ((Array)data).GetValue((i + offset) % length);
              ((Array)data).SetValue(Marshal.PtrToStructure(p, obj.GetType()),i);
              p = (IntPtr)((long)p + ssiz);
              if (n++ >= tdts.dArrayLength) break;
            }
          }
          cc = 0;
          break;
        default:
          {
            int siz = Formats.GetSizeInBytes(fmt);
            IntPtr values = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + fmt), siz, length, offset);
            break;
          }
      }
      return cc;
    }
    public Int32 PutData()
    {
      if (gchRef == null) return 0;
      return PutData(gchRef,(int)tdts.dArrayLength,0);
    }
    public Int32 PutData(Object data)
    {
      int len = 1;
      if (data is Array) len = ((Array)data).GetLength(0);
      return PutData(data, len, 0);
    }
    public Int32 PutData(Object data,int length,int offset)
    {
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);
      switch (fmt)
      {
        case Formats.CF_BYTE:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            cc = native.PutValuesFromByteEx(ref tdts, (byte[])data, length, offset);
          }
          break;
        case Formats.CF_CHAR:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            //char[] values = (char[])data;
            //byte[] values = new byte[length];
            //for (int i = 0; i < length; i++) values[i] = (byte)((char[])data)[i];
            //cc = native.PutValuesFromString(ref tdts, ref values[0], Formats.LFMT_OFFSET + Formats.CF_CHAR, length, offset);
            string values = new String((char[])data);
            cc = native.PutValuesFromString(ref tdts, values, Formats.LFMT_OFFSET+Formats.CF_CHAR, length, offset);
          }
          break;
        case Formats.CF_INT16:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (data is Int16[])
            {
              cc = native.PutValuesFromShortEx(ref tdts, (Int16[])data, length, offset);
            }
            else if (data is UInt16[])
            {
              cc = native.PutValuesFromUnsignedShortEx(ref tdts, (UInt16[])data, length, offset);
            }
          }
          else
          {
            if (data is Int16)
            {
              Int16[] value = {(Int16)data};
              cc = native.PutValuesFromShortEx(ref tdts, value, 1, 0);
            }
            else if (data is UInt16)
            {
              UInt16[] value = {(UInt16)data};
              cc = native.PutValuesFromUnsignedShortEx(ref tdts, value, 1, 0);
            }
          }
          break;
        case Formats.CF_INT32:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (data is Int32[])
            {
              cc = native.PutValuesFromLongEx(ref tdts, (Int32[])data, length, offset);
            }
            else if (data is UInt32[])
            {
              cc = native.PutValuesFromUnsignedLongEx(ref tdts, (UInt32[])data, length, offset);
            }
          }
          else
          {
            if (data is Int32)
            {
              Int32[] value = {(Int32)data};
              cc = native.PutValuesFromLongEx(ref tdts, value, 1, 0);
            }
            else if (data is UInt32)
            {
              UInt32[] value = {(UInt32)data};
              cc = native.PutValuesFromUnsignedLongEx(ref tdts, value, 1, 0);
            }
          }
          break;
        case Formats.CF_INT64:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            IntPtr values = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 1, 8, length, offset);
          }
          break;
        case Formats.CF_FLOAT:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            cc = native.PutValuesFromFloatEx(ref tdts, (float[])data, length, offset);
          }
          else
          {
            float[] value = { (float)data };
            cc = native.PutValuesFromFloatEx(ref tdts, value, 1, 0);
          }
          break;
        case Formats.CF_DOUBLE:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
          //  double[] values = (double[])data;
            cc = native.PutValuesFromDoubleEx(ref tdts, (double[])data, length, offset);
          }
          else
          {
            double[] value = { (double)data };
            cc = native.PutValuesFromDoubleEx(ref tdts, value, 1, 0);
          }
          break;
        case Formats.CF_STRUCT:
          Type dtype = data.GetType();
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            if (dtype.IsArray)
            {
              int n = 0, alen = ((Array)data).Length;
              object obj;
              IntPtr p = tdts.data;
              for (int i = 0; i < alen; i++)
              {
                obj = ((Array)data).GetValue((i + offset) % length);
                Marshal.StructureToPtr(obj, p, true);
                p = (IntPtr)((long)p + ssiz);
                if (n++ >= tdts.dArrayLength) break;
              }
              if (n < length) length = n;
            }
            else
            {
              Marshal.StructureToPtr(data, tdts.data, true);
            }
          }
          cc = 0;
          break;
        default:
          int siz = Formats.GetSizeInBytes(fmt);
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            IntPtr values = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + fmt), 1, siz, length, offset);
          }
          else
          {
            IntPtr value = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, value, (short)(Formats.LFMT_OFFSET + fmt), 1, siz, length, offset);
          }
          break;
      }
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      return cc;
    }
    public Int32 PutData(Object data, int row, int length, int offset)
    {
      int cc = Errors.illegal_format;
      if (!(data is Array)) return Errors.dimension_error;
      if (((Array)data).Rank != 2) return Errors.dimension_error;
      switch (GetTineType(data))
      {
        case Formats.CF_BYTE:
          {
            byte[] values = new byte[length];
            for (int i = 0; i < length; i++) values[i] = ((byte[,])data)[row, i];
            cc = native.PutValuesFromByteEx(ref tdts, values, length, offset);
          }
          break;
        //case Formats.CF_CHAR:
        //  {
        //    char[,] values = (char[,])data;
        //    string s = new String(values[row, 0]);
        //    cc = native.PutValuesFromString(ref tdts, s, Formats.CF_CHAR, length, offset);
        //  }
        //  break;
        case Formats.CF_INT16:
          if (data is Int16[,])
          {
            Int16[] values = new Int16[length];
            for (int i = 0; i < length; i++) values[i] = ((Int16[,])data)[row, i];
            cc = native.PutValuesFromShortEx(ref tdts, values, length, offset);
          }
          else if (data is UInt16[,])
          {
            UInt16[] values = new UInt16[length];
            for (int i = 0; i < length; i++) values[i] = ((UInt16[,])data)[row, i];
            cc = native.PutValuesFromUnsignedShortEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_INT32:
          if (data is Int32[,])
          {
            Int32[] values = new Int32[length];
            for (int i = 0; i < length; i++) values[i] = ((Int32[,])data)[row, i];
            cc = native.PutValuesFromLongEx(ref tdts, values, length, offset);
          }
          else if (data is UInt32[,])
          {
            UInt32[] values = new UInt32[length];
            for (int i = 0; i < length; i++) values[i] = ((UInt32[,])data)[row, i];
            cc = native.PutValuesFromUnsignedLongEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_INT64:
          cc = Errors.not_implemented;
          //{
          //  IntPtr values = GetTineDataPtr(data);
          //  cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 1, 8, length, offset);
          //}
          break;
        case Formats.CF_FLOAT:
          {
            float[] values = new float[length];
            for (int i = 0; i < length; i++) values[i] = ((float[,])data)[row, i];
            cc = native.PutValuesFromFloatEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_DOUBLE:
          {
            double[] values = new double[length];
            for (int i = 0; i < length; i++) values[i] = ((double[,])data)[row, i];
            cc = native.PutValuesFromDoubleEx(ref tdts, values, length, offset);
          }
          break;
        default:
          break;
      }
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      return cc;
    }
    public double GetDataTimeStamp() { return tdts.dTimeStamp; }
    public void SetDataTimeStamp(double timestamp) { tdts.dTimeStamp = timestamp; }
    public Int32 GetSystemStamp() { return tdts.sysStamp; }
    public void SetSystemStamp(Int32 sysStamp) { tdts.sysStamp = sysStamp; }
    public Int32 GetUserStamp() { return tdts.dStamp; }
    public void SetUserStamp(Int32 userStamp) { tdts.dStamp = userStamp; }
    public UInt32 GetDataArrayLength() { return tdts.dArrayLength; }
    public void SetDataArrayLength(UInt32 len) { tdts.dArrayLength = len; }
    public int GetDataFormat() { return tdts.dFormat%256; }
    public string GetDataTag() 
    {
      int n = 0;
      foreach (char c in tdts.dTag) if (c != 0) n++; else break;
      return tdts.dTag != null ? new string(tdts.dTag,0,n) : ""; 
    }
    public void SetDataTag(string tag)
    {
      for (int i=0; i<Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = (char)0;
      if (tag == null) return;
      char[] tagch = tag.ToCharArray();
      for (int i=0; i<Constants.TAG_NAME_SIZE && i<tagch.Length; i++) tdts.dTag[i] = tagch[i];
      return;
    }
    private static List<TTaggedStruct> structLst = new List<TTaggedStruct>();
    public static int RegisterStructure(object tstruct)
    {
      structLst.Add(new TTaggedStruct(tstruct));
      return 0;
    }
  }
  public class native
  {
    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 LaunchRuntimeDebugWindow();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SystemSetCycleTimer();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemInit(Int32 netChk);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemCycle(Int32 chkCmd);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int ExecLinkEx(string devName, string devProperty, ref DTYPE dout, ref DTYPE din, Int16 access, UInt16 timeout);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void TLinkCallback(int id, int cc, IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AttachLinkEx2(string devName,string devProperty,ref DTYPE dout,ref DTYPE din,
         Int16 access, Int32 pollingRate, TLinkCallback cb, Int32 mode, UInt32 cbID, IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void CloseLink(Int32 lnkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataFromLinkId(int linkId, ref DTYPE dout, ref UInt16 status);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetCurrentDataTimeStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetCurrentDataTimeStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDataStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDataStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataSize(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataSizeFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataType(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataTypeFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCurrentLinkStatus(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int FillCompletionStatusString(int linkId, sbyte[] errstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCurrentLinkStatusFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr GetLastLinkError(short cc, sbyte[] errstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr GetSystemUserName();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int _SystemSchedulePropertyEx(string eqm, string prp, Int32 scope);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern UInt32 SetWorkAreaSize(UInt32 size);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 GetFormatSize(Int32 fmt);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate Int32 TEqmFcn(string dev,string prp,ref DTYPE dout,ref DTYPE din,UInt16 acc,IntPtr eqmRef);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void TEqmHook(IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterEquipmentModuleEx(string expName,string eqmName,UInt32 numdevices,
      TEqmFcn fcn, TEqmHook ini, TEqmHook tsk, Int32 rate, TEqmHook exi, IntPtr eqmRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterFecInformation(string name, string sub, string cntxt, 
      string dsc, string loc, string hdw, string resp, UInt16 poff);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetPortOffset(string fecName);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterPropertyInformation(string eqm, string property, ref DTYPE dout, ref DTYPE din,
                                       UInt16 acc, UInt16 atype, UInt16 rowlength, string dsc, Int32 propId, string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RedirectProperty(string eqm,string property,string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterDeviceName(string eqm,string devname,Int32 devnr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RedirectDeviceName(string eqm, string devname, string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetHistoryFilesRepository(string path);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetAppVersion(int major, int minor, int revision);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetAppDate(IntPtr appdate);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetPropertyId(string eqm, string prpName);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDeviceNumberEx(string eqm, string devname, string prpname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDeviceNumber(string eqm, string devname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCallerInformation(string eqm, ref ClnInfo clnInfoList, ref Int32 num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetEqmCompletion(string eqm,string stsstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AddFieldToStruct(string tag, Int32 addr, Int32 size, Int32 fmt, string field);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SealTaggedStruct(string tag, Int32 size, Int32 number);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetStructSize(string tag);
    
    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AcquireAndRegisterStruct(string devName, string tag,int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignNumDevices(string eqm, Int32 num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignDeviceListToProperty(string eqm, string prp, ref NAME64 devlst, Int32 lstlen);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int feclog(string txt);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetFeclogDepth(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetFeclogDepth();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern UInt32 GetWorkAreaSize();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemAssignBufferSpaceEx(UInt32 clnRcvBufferSpace, UInt32 srvRcvBufferSpace, UInt32 clnSndBufferSpace, UInt32 srvSndBufferSpace);

    //[DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    //public static extern void setToThree(ref float v);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int srvlog(string filename, string tag, string txt);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromFloatEx(ref DTYPE d, float[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromDoubleEx(ref DTYPE d, double[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromLongEx(ref DTYPE d, Int32[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromUnsignedLongEx(ref DTYPE d, UInt32[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromShortEx(ref DTYPE d, Int16[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromUnsignedShortEx(ref DTYPE d, UInt16[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromByteEx(ref DTYPE d, byte[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromString(ref DTYPE d, string str, Int32 fmt, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromAny(ref DTYPE d, IntPtr val, Int16 fmt, Int32 sgn, Int32 objSizeInBytes, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsDouble(ref DTYPE d, double[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsFloat(ref DTYPE d, float[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsLong(ref DTYPE d, Int32[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsShort(ref DTYPE d, Int16[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsByte(ref DTYPE d, byte[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsStringEx(ref DTYPE d, char[] val, int fmt, int num, int offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsAny(ref DTYPE d, IntPtr val, Int16 fmt, Int32 objSizeInBytes, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetDataTimeStamp(double ts);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetDataTimeStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double MakeDataTimeStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetSystemDataStamp(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int MakeSystemDataStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetUserStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetAlarmEx(string eqm, Int16 devNr, Int32 almCode, IntPtr almData, byte flags);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceAlarm(string eqm, string devname, Int32 almCode, IntPtr almData, byte flags);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void RemoveDeviceAlarm(string eqm, string devname, Int32 code);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void ClearDeviceAlarm(string eqm, string devname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void ClearAlarm(string eqm, Int16 devNr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendAlarmWatchTable(string eqm, string prp, string dev, Int32 siz, Int32 fmt, Int32 atyp, Int32 sev, Int32 sys, ref ALM_THRESHOLDS thr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendAlarmInfoTable(string eqm, ref ADS ads);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendHistoryInformation(string eqm, string prp, string dev, Int32 len, Int32 fmt, Int32 idx, ref HistorySpecification hspec);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemContexts(IntPtr clist, ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemServers(string context, IntPtr slist, ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemProperties(string server,IntPtr plist,ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDevices(string server,string property,IntPtr dlist,ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetServerAddr(string context, string expName, ref ExpDataStruct srvaddr, ref FecAddrStruct fecaddr);

    private static bool isInitialized = false;
    private static DTYPE nulltype = new DTYPE();
    private static void initializeSettings()
    { // make sure this happens on linux as well !
      if (isInitialized) return;
      nulltype.dFormat = Formats.CF_NULL;
      SystemSetCycleTimer();
      isInitialized = true;
    }
    public static unsafe int ExecLink(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt16 timeout)
    {
      if (!isInitialized) initializeSettings();
      DTYPE p1 = dout != null ? dout.GetDTYPE() : nulltype;
      DTYPE p2 = din != null ? din.GetDTYPE() : nulltype;
      if (din != null) // && din.GetDataFormat() == Formats.CF_STRUCT)
      {
        //din.PutData();
        din.PutData(din.GetDataObject());
      }
      //Int16 acc = (Int16)(access | Access.CA_MUTABLE);
      int cc = (int)ExecLinkEx(devName, devProperty, ref p1, ref p2, access, timeout);
      if (dout != null)
      {
        dout.SetUserStamp(p1.dStamp);
        dout.SetSystemStamp(p1.sysStamp);
        dout.SetDataTimeStamp(p1.dTimeStamp);
        dout.SetDataArrayLength(p1.dArrayLength);
      }
      return cc;
    }
    public static unsafe int AttachLink(string devName, string devProperty, TDataType dout, TDataType din,
      Int16 access, Int32 pollingInterval, TLinkCallback cb, Int32 mode, UInt32 cbID, IntPtr cbRef)
    {
      if (!isInitialized) initializeSettings();
      DTYPE p1 = dout != null ? dout.GetDTYPE() : nulltype;
      DTYPE p2 = din != null ? din.GetDTYPE() : nulltype;
      return (int)AttachLinkEx2(devName, devProperty, ref p1, ref p2, access, pollingInterval, cb, mode, cbID, cbRef);
    }
    native()
    {
      PermissionSet pSet = new PermissionSet(PermissionState.Unrestricted);
      pSet.AddPermission(new SecurityPermission(SecurityPermissionFlag.UnmanagedCode));
      pSet.AddPermission(new FileIOPermission(
          FileIOPermissionAccess.Write |
          FileIOPermissionAccess.Read |
          FileIOPermissionAccess.PathDiscovery, "*.*"));
      initializeSettings();
    }
    private static TLink llnk;
    public static void fixLink(TLink lnk)
    {
      llnk = lnk;
      //GCHandle.Alloc(cb, GCHandleType.Pinned);
    }
  }
  public class TLinkGroup
  {
    private static LinkedList<TLinkGroup> grpTbl = new LinkedList<TLinkGroup>();
    public static TLinkGroup GetGroup(TLink.TLinkCallback lcb)
    {
      foreach (TLinkGroup g in grpTbl)
      {
        if (g.lcb == lcb) return g;
      }
      TLinkGroup gn = new TLinkGroup(lcb);
      grpTbl.AddLast(gn);
      return gn;
    }
    public TLinkGroup(TLink.TLinkCallback lcb)
    {
      this.lcb = lcb;
      lcbdg = lclCb;
    }
    private LinkedList<TLink> members = new LinkedList<TLink>();
    public void AddMember(TLink lnk)
    {
      lock (members)
      {
        if (members.Contains(lnk)) return;
        members.AddLast(lnk);
      }
    }
    public void RemoveMember(TLink lnk)
    {
      lock (members)
      {
        if (!members.Contains(lnk)) return;
        members.Remove(lnk);
      }
    }
    private TLink.TLinkCallback lcb;
    public Boolean HasGroupCallback(TLink.TLinkCallback gcb)
    {
      if (lcb == null) return false;
      return gcb == lcb;
    }
    public TLink[] GetMembers()
    {
      lock (members)
      {
        if (members.Count == 0) return null;
        TLink[] mems = new TLink[members.Count];
        members.CopyTo(mems, 0);
        return mems;
      }
    }
    private Int32 lnkStatus;
    /// <summary>
    /// The current link status as an integer value.
    /// </summary>
    /// <returns>The current link status.</returns>
    public Int32 GetStatus() { return lnkStatus; }
    private Int32 cbId = 0; // for those who insist
    public Int32 GetCallbackId() { return cbId; }
    private IntPtr cbRef = IntPtr.Zero;  // for those who insist
    public IntPtr GetCallbackReference() { return cbRef; }
    private Int32 lnkId = -1;
    public Int32 GetLinkId() { return lnkId; }
    public native.TLinkCallback lcbdg;
    public unsafe void lclCb(Int32 id, Int32 cc, IntPtr cbRef)
    { // will only be called from below when they're all there !
      this.lnkStatus = cc;
      this.cbId = id;
      this.cbRef = cbRef;
      lock (members)
      {
        foreach (TLink m in members)
        {
          m.SetStatus(native.GetCurrentLinkStatus(m.GetLinkId()));
          m.memupd();
        }
        if (lcb != null && members.Count > 0) lcb(members.First.Value);
      }
    }
  }
  /// <summary>
  /// Defines a connection between the local client and
  /// the remote server's targeted end point
  /// </summary>
  /// <remarks>
  /// TLink is a TINE Link object which defines a connection between the local client and
  /// the remote server's targeted end point. Data may be exchanged either synchronously
  /// (via the Execute() method) or asynchronously (via the Attach() method).  Simple 
  /// commands which do not involve data can likewise be issued synchronously.
  /// </remarks>
  public class TLink
  {
    public Boolean IsGrouped()
    {
      return grp == null ? false : true;
    }
    public Boolean Equals(TLink lnk)
    {
      return lnk == this;
    }
    private sbyte[] lnkStatusString = new sbyte[Constants.STATUS_SIZE];
    private String lnkDevName;
    public String GetDeviceName() { return lnkDevName; }
    private String lnkPrpName;
    public String GetProperty() { return lnkPrpName; }
    public String GetKey() 
    {
      if (lnkDevName == null || lnkPrpName == null) return "[empty]";
      return lnkDevName + "[" + lnkPrpName + "]"; 
    }
    private TDataType lnkDout;
    public TDataType GetOutputDataType() { return lnkDout; }
    private TDataType lnkDin;
    public TDataType GetInputDataType() { return lnkDin; }
    private UInt16 lnkAccess;
    public UInt16 GetAccess() { return lnkAccess; }
    private Int32 lnkStatus;
    /// <summary>
    /// The current link status as an integer value.
    /// </summary>
    /// <returns>The current link status.</returns>
    public Int32 GetStatus() { return lnkStatus; }
    public void SetStatus(Int32 status) { lnkStatus = status; }
    private Int32 lnkAccessRate = 1000;
    public Int32 GetAccessRate() { return lnkAccessRate; }
    private UInt32 lnkAccessMode;
    public UInt32 GetAccessMode() { return lnkAccessMode; }
    private UInt32 cbId = 0; // for those who insist
    public Int32 GetCallbackId() { return (Int32)cbId; }
    private IntPtr cbRef = IntPtr.Zero;  // for those who insist
    public IntPtr GetCallbackReference() { return cbRef; }
    private TLinkGroup grp = null;
    public TLinkGroup GetGroup()
    {
      return grp;
    }
    public void DetachFromGroup()
    {
      if (grp != null) grp.RemoveMember(this);
      grp = null;
    }
    private Int32 lnkId = -1;
    public Int32 GetLinkId() { return lnkId; }
    public delegate void TLinkCallback(TLink lnk);
    private TLinkCallback lcb;
    private native.TLinkCallback lnkCb;
    private int _execute(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt32 timeout)
    {
      int cc = native.ExecLink(devName, devProperty, dout, din, access, (UInt16)timeout);
      Object dobj;
      if (dout != null && (dobj = dout.GetReferencedDataObject()) != null)
      {
        dout.GetData(dobj);
      }
      lnkStatus = cc;
      return cc;
    }
    public int Execute(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt32 timeout)
    {
      return _execute(devName, devProperty, dout, din, access, (UInt16)timeout);
    }
    /// <summary>
    /// Executes a synchronous link.
    /// </summary>
    /// <remarks>
    /// Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
    /// a timeout has ensued. This call uses the default timeout of 1000 msec.
    /// </remarks>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    /// \b Example:
    /// \include eg_Execute.cs
    public int Execute()
    {
      return _execute(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, (UInt16)1000);
    }
    /// <summary>
    /// Executes a synchronous link.
    /// </summary>
    /// <remarks>
    /// Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
    /// a timeout has ensued.
    /// </remarks>
    /// <param name="timeout">is the designated timeout in milliseconds.</param>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    public int Execute(UInt32 timeout)
    {
      return _execute(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, (UInt16)timeout);
    }
    public int GetCompletionDataSize()
    {
      return native.GetCompletionDataSize(lnkId);
    }
    public double GetCompletionDataTimeStamp()
    {
      return native.GetCurrentDataTimeStamp(lnkId);
    }
    public double GetCompletionDataStamp()
    {
      return native.GetDataStamp(lnkId);
    }
    public double GetCompletionSystemStamp()
    {
      return native.GetSystemDataStamp(lnkId);
    }
    public unsafe void memupd()
    {
      if (lnkDout.GetDataFormat() != Formats.CF_STRUCT) return;
      IntPtr ptr = lnkDout.GetDTYPE().data;
      object rdo = lnkDout.GetReferencedDataObject();
      if (!rdo.GetType().IsArray) return;
      Array tts = (Array)rdo;
      Type t = tts.GetValue(0).GetType();
      int slen = native.GetStructSize(lnkDout.GetDataTag());
      for (int i = 0; i < (int)lnkDout.GetDataArrayLength(); i++)
      {
        ptr = new IntPtr(ptr.ToInt64() + (i * slen));
        object o = Marshal.PtrToStructure(ptr, t);
        tts.SetValue(o, i);
      }
    }
    public unsafe void lclCb(Int32 id, Int32 cc, IntPtr cbRef)
    {
      this.lnkStatus = cc;
      this.cbId = (UInt32)id;
      this.cbRef = cbRef;
      memupd();
      if (lcb != null) lcb(this);
    }
    public int Attach(string devName, string devProperty, TDataType dout, TDataType din, UInt16 access, Int32 accessRate, TLinkCallback cb, UInt32 mode, UInt32 cbId, IntPtr cbRef)
    {
      lnkDevName = devName;
      lnkPrpName = devProperty;
      lnkDout = dout;
      lnkDin = din;
      lnkAccess = access;
      this.cbId = cbId;
      this.cbRef = cbRef;
      return Attach(mode, cb, accessRate);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// This method automatically choses to monitor the endpoint in CM_TIMER mode at an access rate
    /// of 1 Hz (1000 msec polling interval).
    /// </remarks>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    public int Attach(TLinkCallback cb)
    {
      return Attach(Modes.CM_TIMER, cb, 1000);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// This method uses an access rate of 1 Hz (1000 msec polling interval).
    /// </remarks>
    /// <param name="mode">is the desired access mode, i.e. CM_TIMER, CM_DATACHANGE, CM_EVENT, etc. with
    /// possible modifiers (see Modes).</param>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    public int Attach(UInt32 mode, TLinkCallback cb)
    {
      return Attach(mode, cb, 1000);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// </remarks>
    /// <param name="mode">is the desired access mode, i.e. CM_TIMER, CM_DATACHANGE, CM_EVENT, etc. with
    /// possible modifiers (see Modes).</param>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <param name="accessRate">is the desired access polling interval in milliseconds. This interval is
    /// maintained at the server and is used to trigger the server's scheduler.</param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    /// \b Example:
    /// \include eg_Attach.cs
    public int Attach(UInt32 mode, TLinkCallback cb, Int32 accessRate)
    {
      if (lnkDout != null && lnkDout.isByRef) throw new Exception("invalid output data reference: asynconchronous link on non-array type not allowed");
      lcb = cb;
      lnkCb = lclCb;
      native.fixLink(this);
      if ((mode & Modes.CM_GROUPED) == Modes.CM_GROUPED)
      {
        TLinkGroup g = TLinkGroup.GetGroup(cb);
        if (g != null)
        {
          g.AddMember(this);
          grp = g;
          lnkCb = g.lcbdg; // g.lclCb;
        }
        else
        {
          UInt32 mask = Modes.CM_GROUPED;
          mode &= (~mask);
        }
      }
      lnkAccessMode = mode;
      lnkAccessRate = accessRate;
      lnkId = native.AttachLink(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, accessRate, lnkCb, (Int32)mode, cbId, cbRef);
      return lnkId;
    }
    /// <summary>
    /// Closes a TLink.
    /// </summary>
    /// <remarks>
    /// If the link is not currently attached to an input, this call has no effect.
    /// </remarks>
    /// <returns>0 upon success or a TINE error code.</returns>
    public int Close()
    {
      if (lnkId < 0) return -1;
      if (grp != null) DetachFromGroup();
      native.CloseLink(lnkId);
      return 0;
    }
    /// <summary>
    /// The current link status as a string
    /// </summary>
    /// <returns>The current link status as a string.</returns>
    public string GetStatusString()
    {
      //native.GetLastLinkError((short)lnkStatus, lnkStatusString);
      native.FillCompletionStatusString(lnkId, lnkStatusString);
      char[] errchr = new char[Constants.STATUS_SIZE];
      int i;
      for (i = 0; i < Constants.STATUS_SIZE && lnkStatusString[i] != 0; i++) errchr[i] = (char)lnkStatusString[i];
      return new String(errchr,0,i);
    }
    public TLink()
    {
      cbId = (UInt32)this.GetHashCode();
    }
    /// <summary>
    /// Establishes a link endpoint to the targeted server.
    /// </summary>
    /// <remarks>
    /// Creating a TLink instance will only register the desired link
    /// parameters at the client side.  It will NOT exchange data or 
    /// otherwise establish a connection.
    /// 
    /// Note that supplying an input data object does NOT constitute a
    /// 'WRITE' command to the targeted server!  Input data is simply 'input
    /// data'.  The access of the call is determined by the the access parameter.
    /// </remarks>
    /// <param name="devName">is the desired device name of the target, in the form of
    /// '/context/server/device'.
    /// </param>
    /// <param name="devProperty">is the desired property of the target</param>
    /// <param name="dout">is the desired output data object (data returned from the server)</param>
    /// <param name="din">is the given input data object (data sent to the server)</param>
    /// <param name="access">is the desired access of this link object</param>
    public TLink(string devName, string devProperty, TDataType dout, TDataType din, UInt16 access)
    {
      lnkDevName = devName;
      lnkPrpName = devProperty;
      lnkDout = dout;
      lnkDin = din;
      lnkAccess = access;
      cbId = (UInt32)this.GetHashCode();
    }
    ~TLink()
    { // the Link object is now a null pointer !
      Close(); // just in case ...
      //Console.WriteLine("link destructor is being called");
      GC.Collect();
    }
  }
  public class TServerAddress
  {
    private string srv;
    private string ctx;
    private string eqm;
    private string fec;
    private IPAddress ip;
    private Int32 po;
    public TServerAddress(string server, string context, string eqmName, string fecName, IPAddress ipAddr, Int32 portOffset)
    {
      srv = server == null ? "" : new string(server.ToCharArray());
      ctx = context == null ? "" : new string(context.ToCharArray());
      eqm = eqmName == null ? "" : new string(eqmName.ToCharArray());
      fec = fecName == null ? "" : new string(fecName.ToCharArray());
      ip = ipAddr;
      po = portOffset;
    }
    public TServerAddress(char[] server, char[] context, char[] eqmName, char[] fecName, IPAddress ipAddr, Int32 portOffset)
    {
      srv = server == null ? "" : new string(server);
      ctx = context == null ? "" : new string(context);
      eqm = eqmName == null ? "" : new string(eqmName);
      fec = fecName == null ? "" : new string(fecName);
      ip = ipAddr;
      po = portOffset;
    }
    public string GetServer() { return srv; }
    public void SetServer(string serverName) 
    { 
      srv = serverName == null ? "" : new string(serverName.ToCharArray()); 
    }
    public string GetContext() { return ctx; }
    public void SetContext(string context) 
    {
      ctx = context == null ? "" : new string(context.ToCharArray()); 
    }
    public string GetEqm() { return eqm; }
    public string GetFec() { return fec; }
    public IPAddress GetIpAddress() { return ip; }
    public Int32 GetPortOffset() { return po; }
  }
  public class TQuery
  {
    /// <summary>
    /// Retrieves a list of server contexts via query to the Equipment Name Server. 
    /// </summary>
    /// This call retrieves a list of all server context managed by the Equipment Name Server
    /// <returns>A string array containing the desired context list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemContexts()
    {
      NAME32[] clst = new NAME32[100];
      Int32 n = 100;
      IntPtr p = TDataType.GetDataPtr(clst);
      int cc = native.GetSystemContexts(p, ref n);
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      String[] ctxs = new String[n];
      for (int i = 0; i < n; i++) ctxs[i] = clst[i].GetName();
      return ctxs;
    }
    /// <summary>
    /// Retrieves a list of device servers via query to the Equipment Name Server. 
    /// </summary>
    /// This call retrieves a list of all Front End Computer Names (FECs) associated 
    /// with the server context given. In addition, the context string can carry a 
    /// targeted subsystem for a more specific list.
    /// <param name="context">is the context for which the query is to be made. A narrower 
    /// query can also be made, where only servers belonging to a specific subsystem are 
    /// returned. If this is desired, then pass a context string of the form 
    /// 'context/subsystem' (e.g. "PETRA/DIAG"). </param>
    /// <returns>A string array containing the desired server list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemServers(String context)
    {
      NAME32[] lst = new NAME32[1024];
      Int32 n = 1024;
      IntPtr p = TDataType.GetDataPtr(lst);
      int cc = native.GetSystemServers(context, p, ref n);
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      String[] srvs = new String[n];
      for (int i = 0; i < n; i++) srvs[i] = lst[i].GetName();
      return srvs;
    }
    /// <summary>
    /// Retrieves a list of properties via query to the server given. 
    /// </summary>
    /// This call retrieves a list of all properties registered on the server 
    /// associated with full device server name specified. A classic device server 
    /// will export its behavior through its properties, and export its "instances" 
    /// through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device 
    /// found on this server". In this case, simply specifying the device server name 
    /// as the first input parameter will return the set of properties valid for all 
    /// devices registered on the server. Under some circumstances, however, the set 
    /// of properties associated with a particular device might be different from device 
    /// to device. If this is the case, the full device server name should be used as the 
    /// first parameter, where the target device name is also sent as part of the query.
    /// <param name="server">is the full device server name for which the property query is to be made </param>
    /// <returns>A string array containing the desired property list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemProperties(String server)
    {
      NAME64[] lst = new NAME64[2048];
      Int32 n = 2048;
      IntPtr p = TDataType.GetDataPtr(lst);
      int cc = native.GetSystemProperties(server, p, ref n);
      if ((cc & Errors.CE_SENDDATA) != 0) cc = 0;
      if (cc != 0) throw new Exception("query failed: <"+cc+"> "+Errors.ToString(cc));
      if (n <= 0) return null;
      String[] prps = new String[n];
      for (int i = 0; i < n; i++) prps[i] = lst[i].GetName();
      return prps;
    }
    /// <summary>
    /// Retrieves a list of device names associated with a given property via query to the server specified. 
    /// </summary>
    /// <remarks>
    /// This call retrieves a list of all device names registered on the server associated with the 
    /// property specified in the call. A classic device server will export its behavior through its properties, 
    /// and export its "instances" through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device found on this server". 
    /// For certain classes of server (in particular middle layer servers) the properties exported might refer 
    /// to 'apples' and 'oranges' simply because the server is archiving them. Or there might be other reasons why 
    /// the device list associated with a particular property has little or nothing to do with that associated with 
    /// some other property. Use this call to obtain the device list associated with a particular property.
    /// </remarks>
    /// <param name="server"> is the full device server name for which the property query is to be made </param>
    /// <param name="property">is the property for which the device list is desired. </param>
    /// 
    /// <returns>A string array containing the desired device list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemDevices(String server,String property)
    {
      NAME64[] lst = new NAME64[2048];
      int n = 2048;
      IntPtr p = TDataType.GetDataPtr(lst);
      int cc = native.GetSystemDevices(server, property, p, ref n);
      if ((cc & Errors.CE_SENDDATA) != 0) cc = 0;
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      String[] devs = new String[n];
      for (int i = 0; i < n; i++) devs[i] = lst[i].GetName();
      return devs;
    }
    public static int GetPropertyInformation(string dev, string prp, ref PrpQueryStruct[] pqs, ref int num)
    {
      TLink lnk;
      TDataType dout, din, dinptr = null;
      DTYPE dtout;
      int cc = 0;
      short n, nprp = 0;
      PrpQueryStruct[] prpqs = null;

      if (dev == null || pqs == null) return Errors.invalid_parameter;
      dout = new TDataType(ref nprp); // make 'a' call in order to be sure the client cycle is initialized
      lnk = new TLink(dev, "NPROPERTIES", dout, null, Access.CA_READ);
      if ((cc = lnk.Execute(500)) != 0) return cc;
      if ((n = (short)num) <= 0) n = prp == null ? nprp : (short)10;
      if ((prpqs = pqs) == null)
      { /* caller wants me to allocate this */
        prpqs = new PrpQueryStruct[n];
      }
      if (prp != null)
      {
        NAME64[] n64 = new NAME64[1];
        n64[0] = new NAME64();
        n64[0].SetName(prp);
        din = new TDataType(n64);
        dinptr = din;
      }
      dtout = new DTYPE();
      dtout.dArrayLength = (uint)n;
      dtout.dFormat = Formats.LFMT_OFFSET + Formats.CF_STRUCT;
      dtout.data = TDataType.GetDataPtr(prpqs, "PRPQSr4");
      dtout.dTag = "PRPQSr4".ToCharArray();
      dout = new TDataType(dtout,prpqs);
      lnk = new TLink(dev, "PROPERTIES", dout, dinptr, Access.CA_READ);
      if ((cc = lnk.Execute(500)) != 0) return cc;
      if (cc == 0 || (cc & Errors.CE_SENDDATA) != 0)
      { /* modern server understood the query */
        num = (int)dout.GetDataArrayLength();
        pqs = prpqs;
        return 0;
      }
      return cc;
    }
    public static TServerAddress GetServerAddress(string context, string server)
    {
      TServerAddress sa = null;
      ExpDataStruct ea = new ExpDataStruct();
      FecAddrStruct fa = new FecAddrStruct();
      int cc = 0;
      if ((cc=native.GetServerAddr(context, server, ref ea, ref fa)) == 0)
      { // TODO: ipv6 branch (the IPAddress constructor makes a decision based on the length of the byte array!)
        byte[] ipb = new byte[4];
        for (int i = 0; i < 4; i++) ipb[i] = fa.netAdr[i];
        IPAddress ip = new IPAddress(ipb);
        sa = new TServerAddress(ea.EqmContext,ea.ExportName,ea.EqmName,fa.fecName,ip,fa.portOffset);
      }

      return sa;
    }
    public static int AcquireAndRegisterStruct(string devName,string tag,int number)
    {
      return native.AcquireAndRegisterStruct(devName, tag, number);
    }
  }
  /// <summary>
  /// TKernel is the TINE Kernel (singleton) class, whose static methods are used to manage 
  /// the TINE engine.
  /// </summary>
  public class TKernel
  {
    /// <summary>
    /// Sets the server-side work area size which is coupled to the maximum transport size for 
    /// supported by a single-threaded server.
    /// </summary>
    /// <remarks>
    /// Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
    /// of the caller's requested data size, the other buffer being in the case of single threaded servers,
    /// a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
    /// size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
    /// servers will dynamically allocate both buffers on a call by call basis.
    /// </remarks>
    /// <param name="sizeInBytes">is the size in bytes to be reserved for the server's work area and hence the maximum
    /// allowed transport size for any property supported by the server.
    /// (default: 64 KBytes on most operating systems).</param>
    /// <returns>The current setting for the server's work area.</returns>
    public static UInt32 SetWorkAreaSize(UInt32 sizeInBytes)
    {
      return native.SetWorkAreaSize(sizeInBytes);
    }
    /// <summary>
    /// Gets the server-side work area size which is coupled to the maximum transport size for 
    /// supported by a single-threaded server.
    /// </summary>
    /// <remarks>
    /// Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
    /// of the caller's requested data size, the other buffer being in the case of single threaded servers,
    /// a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
    /// size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
    /// servers will dynamically allocate both buffers on a call by call basis.
    /// </remarks>
    /// <returns>The current setting for the server's work area.</returns>
    public static UInt32 GetWorkAreaSize()
    {
      return native.GetWorkAreaSize();
    }
    /// <summary>
    /// Assigns specific buffer capacity for certain configurable TINE sockets.
    /// </summary>
    /// <remarks>
    ///The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a 
    ///reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
    ///performance reason, more or less buffer capactiy is needed, this routine should be called prior to
    ///any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large 
    ///amount of network resources might want to set these values smaller than the defaults.  Video system
    ///clients expecting extreme amounts of incoming data might want to set these values to a much larger
    ///amount.
    ///
    ///This call allows the setting of the server side receive buffer space as well as the client side
    ///send buffer space. This is potentially applicable in rare cases (middle layer servers) where large
    ///data payloads are sent as input from a client to a server.
    ///
    ///As in the case of _SystemAssignBufferSpace(), the UDP specific send buffer size is kept to a reasonable
    ///value, no larger than twice the registered packet MTU.  UDP datagram transmission is not inhibited by
    ///any client-server flow control which might cause the send buffer space to stack up.
    /// </remarks>
    /// <param name="clnRcvBufferSpace">is the amount of receive buffer space (in bytes) to assign to the configurable 
    /// TINE client sockets. A value of '0' will assign the default receive buffer capacity.</param>
    /// <param name="srvRcvBufferSpace">is the amount of receive buffer space (in bytes) to assign to the configurable TINE
    /// server sockets. A value of '0' will assign the same value as clnRcvBufferSpace (and damping this value by a
    /// factor of 2 should it exceed the default receive buffer capacity).</param>
    /// <param name="clnSndBufferSpace">s the amount of send buffer space (in bytes) to assign to the configurable TINE 
    /// client sockets. A value of '0' will assign the same value as srvSndBufferSpace (and damping this value by a 
    /// factor of 2 should it exceed the default receive buffer capacity).</param>
    /// <param name="srvSndBufferSpace">is the amount of send buffer space (in bytes) to assign to the configurable  
    /// TINE server TCP or STREAM sockets. 
    /// The UDP socket buffer space will be at least as large as twice the assigned UDP packet MTU.
    /// A value of 0' will assign the default send buffer capacity.</param>
    /// 
    /// <returns>0 on success or a TINE error code</returns>
    public static Int32 AssignBufferSpace(UInt32 clnRcvBufferSpace, UInt32 srvRcvBufferSpace, UInt32 clnSndBufferSpace, UInt32 srvSndBufferSpace)
    {
      return native._SystemAssignBufferSpaceEx(clnRcvBufferSpace,srvRcvBufferSpace,clnSndBufferSpace,srvSndBufferSpace);
    }
    /// <summary>
    /// Assigns specific buffer capacity for certain configurable TINE sockets.
    /// </summary>
    /// <remarks>
    /// The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a 
    /// reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
    /// performance reason, more or less buffer capactiy is needed, this routine should be called prior to
    /// any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large 
    /// amount of network resources might want to set these values smaller than the defaults.  Video system
    /// clients expecting extreme amounts of incoming data might want to set these values to a much larger
    /// amount.
    ///
    /// The principal benificiaries of the 'receive buffer space' are the client sockets (UDP,TCP and SREAM), whereas
    /// the principal benificiaries of the 'send buffer space' are the server sockets (TCP and STREAM). The
    /// remaining socket buffer space (e.g. client-side send buffer space or server side receive buffer space)
    /// will be held at 'reasonable' values, the assumption being that the dominant data delivery will generally
    /// server (sending) to client (receiving).
    /// </remarks>
    /// <param name="rcvBufferSpace">s the amount of receive buffer space (in bytes) to assign to the configurable TINE sockets.
    /// A value of '0' will assign the default receive buffer capacity.</param>
    /// <param name="sndBufferSpace">sndBufferSpace is the amount of send buffer space (in bytes) to assign to the configurable TINE TCP or  
    /// STREAM sockets.  The UDP socket send buffer space if not influenced by this parameter and 
    /// will be at least as large as twice the assigned UDP packet MTU. See _SystemAssignUdpSendBufferSpace().
    /// A value of 0' will assign the default send buffer capacity.</param>
    /// <returns>0 on success or a TINE error code</returns>
    public static Int32 AssignBufferSpace(UInt32 rcvBufferSpace, UInt32 sndBufferSpace)
    {
      return native._SystemAssignBufferSpaceEx(rcvBufferSpace, 0, 0, sndBufferSpace);
    }
    /// <summary>
    /// Appends the given text to the FEC process's log file.
    /// </summary>
    /// <remarks>
    /// Each log entry is prefixed by the FEC process name (i.e. the 'FECNAME') and the time
    /// and date.  The location of the log file (named 'fec.log' and rotated once into 'fec.bak')
    /// is determined by the environment variable FEC_LOG (first choice) or FEC_HOME. If
    /// neither environment variable is set, then the fec.log file will be written to the
    /// current working directory.
    /// </remarks>
    /// <param name="text">is the desired text to be appended to the FEC log file</param>
    /// <returns>0 upon success or a TINE error code (e.g. file_error)</returns>
    public static Int32 FecLog(string text)
    {
      return native.feclog(text);
    }
    /// <summary>
    /// Sets the approximate maximum number of log entries in the FEC process's log file.
    /// </summary>
    /// <remarks>
    /// The number of long entries is approximately deterimed by assuming the the average
    /// line of text contains 160 characters.  If the number so deterimed exceeds the
    /// given FEC log depth, the current log file 'fec.log' will be rotated into
    /// 'fec.bak'.
    /// </remarks>
    /// <param name="depth">
    /// is the maximum number of lines the FEC's log file is allowed to have.
    /// (default = 1000)
    /// </param>
    public static void SetFecLogDepth(Int32 depth)
    {
      native.SetFeclogDepth(depth);
    }
    /// <summary>
    /// Retrieves the current setting for the maximum number of log entries in the FEC process's log file.
    /// </summary>
    /// <returns>The current setting.</returns>
    public static Int32 GetFecLogDepth()
    {
      return native.GetFeclogDepth();
    }
    /// <summary>
    /// Appends the given text to the designated log file.
    /// </summary>
    /// <param name="filename">
    /// is the file name to which the log entry is to be appended.  The file will be created in the
    /// 'FEC_LOG' or 'FEC_HOME' directory (as in the case of fec.log), and rotated using the same
    /// log depth established via SetFecLogDepth().
    /// </param>
    /// <param name="tag">
    /// is a 'procedure tag' which will prefix the given log entry.  If ommitted, the tag 'logger' will
    /// be used.
    /// </param>
    /// <param name="text">
    /// is the desired log entry text.
    /// </param>
    /// <returns>
    /// 0 upon success or a TINE error code.
    /// </returns>
    public static Int32 SrvLog(string filename, string tag, string text)
    {
      return native.srvlog(filename,tag,text);
    }
    /// <summary>
    /// Initializes the TINE server engine.
    /// </summary>
    /// <remarks>
    /// This method MUST be invoked for the TINE server engine to function property.
    /// Should be called prior to the start of the main TINE server engine. Any deviations from default system parameters should be made before 
    /// server initialization. This might include server settings which fix the size of the local connection table entries, the maximum date transport 
    /// size, etc. Note that InitializeServer() will begin allocating space for its connection tables, so modifying these values away from their defaults 
    /// will have no effect after InitializeServer() has been called. Also, if the FEC Name is given by calling RegisterFecName() then this should also 
    /// preceed a call to InitializeServer().
    /// </remarks>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    public static Int32 InitializeServer()
    {
      return native._SystemInit(1);
    }
    /// <summary>
    /// Required call for the TINE engine to function property.
    /// </summary>
    /// <remarks>
    /// On a multithreaded platform (e.g. all .NET implementations) loops around  _SystemCycle().
    /// In a single thread environment,e.g., MSDOS blocks the main process.
    /// </remarks>
    public static void StartServices()
    {
      native.SystemSetCycleTimer();
    }
    /// <summary>
    /// Sets the intrinsic data timestamp to the value given. 
    /// </summary>
    /// <remarks>
    /// This call is typically used inside an equipment 
    /// function to set the data timestamp associated with the request being made. For instance, when the data 
    /// are taken from the hardware IO loop, a call to makeDataTimeStamp() is made to get the current system timestamp. 
    /// Then when a request for data is made the caller can know the precise time of data acquisition. 
    /// If this routine is not called, then the current system time will be used to stamp the data.
    /// </remarks>
    /// <param name="timestamp">The timestamp to be associated with the next data set.</param>
    public static void SetDataTimeStamp(double timestamp)
    {
      native.SetDataTimeStamp(timestamp);
    }
    /// <summary>
    /// Provides the last established data timestamp.
    /// </summary>
    /// <remarks>
    /// A call to getDataTimeStamp() immediately following a call to ExecLink() or within a 
    /// callback routine will yield the data timestamp of the call.
    /// Note: This call is not thread-safe. Synchronous data acquisition on two different threads 
    /// could conceivably interfer with the value returned from this routine. A thread-safe way of 
    /// obtaining the data timestamp associated with a call is to use the .dTimestamp field of the DTYPE 
    /// object for synchronous calls such as ExecLink() or ExecLinkEx() or to use getCurrentDataTimeStampFromCallbackId() 
    /// or getCurrentDataTimeStamp() (which take a link identifier as input parameter) for asynchronous links.
    /// </remarks>
    /// <returns>The data timestamp currently in place.</returns>
    public static double GetDataTimeStamp()
    {
      return native.GetDataTimeStamp();
    }
    /// <summary>
    /// Makes a data timestamp according to the current system time.
    /// </summary>
    /// <returns>A tine data timestamp.</returns>
    public static double MakeDataTimeStamp()
    {
      return native.MakeDataTimeStamp();
    }
    /// <summary>
    /// Provides the last established system data stamp
    /// </summary>
    /// <remarks>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.
    /// </remarks>
    /// <returns>The system data stamp current in place.</returns>
    public static int GetSystemDataStamp()
    {
      return native.GetSystemDataStamp(-1);
    }
    /// <summary>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.  This call 
    /// is in parallel with MakeDataTimeStamp() and is nominally equivalent to GetSystemDataStamp.
    /// </summary>
    /// <returns>The system data stamp current in place.</returns>
    public static int MakeSystemDataStamp()
    {
      return native.MakeSystemDataStamp();
    }
    /// <summary>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.  A server may need to
    /// record the current system stamp at the time of e.g. a hardware trigger and apply the value in
    /// subsequent calls to access data.  This method can be used toward these ends.
    /// </summary>
    /// <param name="value">
    /// is the value of the system data stamp to apply to current data access.
    /// </param>
    public static void SetSystemDataStamp(int value)
    {
      native.SetSystemDataStamp(value);
    }
    /// <summary>
    /// Assigns a FEC Name and descriptive information to the server process. 
    /// </summary>
    /// <remarks>
    /// Servers must be assigned to a Front End Computer (FEC), which for many operating systems corresponds 
    /// to a process running on the computer rather than the computer itself (meaning there can in some cases 
    /// be more than one servers with different 'FEC' names running on the same computer. There can be several 
    /// device servers attached to the same FEC (i.e. sharing the same address space), so the FEC name is a distinct 
    /// quantity from the device server name (see RegisterExport()), and this name must also be system-wide unique. 
    /// The call RegisterFecInformation() supercedes RegisterFecNameEx() and RegisterFecName() and differs from them 
    /// in that it accepts 'Context' and 'SubSystem' as parameters, both of which will be applied to all device 
    /// servers attaching to the same FEC name.
    /// The FEC name can also be (and frequently is) registered via the presence of an 'fecid.csv' startup database file.
    /// </remarks>
    /// <param name="fecName">The FEC name indentifying the Front End Computer and to which all registered equipment modules are bound.</param>
    /// <param name="defaultSubSystem">The subsystem to which attached device servers belong. Note that the subsystem (unlike context) is itself 
    /// not part of the name space, although the name space can be queried for device servers belonging to a subsystem. 
    /// This information is now deduced automatically from the library build.</param>
    /// <param name="defaultContext">A 32-character string giving the context of all registered equipment modules. 
    /// (earlier versions of TINE truncated at 16 characters).</param>
    /// <param name="fecDescription">A 64-character desciption of the FECs server duties. A 'subsystem' can be associated with ALL equipment modules 
    /// found on the FEC by prefixing the description with a subsystem tag. This information will be parsed by the Equipment name 
    /// server and used for sub-system specific queries. For example, the associate the FEC "HEPHF" with the subsystem "RF", the description field 
    /// might be "[RF]Hera Proton HF Control".</param>
    /// <param name="fecLocation">A 32-character string giving the phyical location of the FEC.</param>
    /// <param name="attachedHardware">A 32-character brief description of the IO hardware found on the FEC.</param>
    /// <param name="responsible">A 32-character string listing the developer(s) responsible for the FEC. 
    /// Note: The Equipment Name Server will allow the removal of FECs and associated equipment modules to the user(s) specified here.</param>
    /// <param name="portOffset">The 'Port Offset' to be applied to the FEC. This parameter plays an important role where more that 
    /// one FEC is to run on a machine running an operating system using virtual memory. In such cases 'Front End Computer' is a misnomer, 
    /// since "FEC" really refers to a process running on the computer to which one or more equipment modules are bound. For operating systems 
    /// where all processes run in the same address space (such as VxWorks, MS-DOS, Win16, NIOS) there is in fact only one such processes managing 
    /// all requistered equipment modules. For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.), you can have many such 
    /// processes running independently of one another. Since each such process must listen on a unique set of server ports, you must see to it that 
    /// all such FEC processes are registered with a unique 'Port Offset'.</param>
    /// <returns>0 if successful, otherwise a TINE completion code which can be interpreted by a call to GetLastLinkError()</returns>
    public static Int32 RegisterFecInformation(String fecName,String defaultSubSystem,String defaultContext,String fecDescription,String fecLocation,String attachedHardware,String responsible,UInt16 portOffset)
    {
      return native.RegisterFecInformation(fecName, defaultSubSystem, defaultContext, fecDescription, fecLocation, attachedHardware, responsible, portOffset);
    }
    /// <summary>
    /// Obtains the FEC port offset appropriate for the give FEC name.
    /// </summary>
    /// <remarks>
    /// In cases where the pure registration API routines are used to register FEC and server information (as opposed to a common fecid.csv 
    /// configuration file) it is often a more cumbersome task to ensure a unique FEC port on the host machine. This routine can be used to 
    /// this end. It will search the existing FEC manifest on the current host in order to find the matching entry for the given FEC name. 
    /// If the entry is not found, then the 'next free port' is returned. In this way the FEC can always register with its own proper FEC port, 
    /// without the need to hard code this number.
    /// 
    /// If this routine is called after the server is running and the FEC name is not found in the manifest the -name_unknown is returned.
    ///
    /// </remarks>
    /// <param name="fecName"></param>
    /// <returns>
    /// the port offset applicable for the given FEC name or the a negative number giving the negative of the TINE return code indicating an error.
    /// </returns>
    public static Int32 GetPortOffset(String fecName)
    {
      return native.GetPortOffset(fecName);
    }
    /// <summary>
    /// Assigns pertinent information for the specified property. 
    /// </summary>
    /// <remarks>
    /// Servers export their behavior through Properties, which can be thought of as 'properties' 
    /// in the Object Oriented sense with Get and Set methods. That is, any read or write access 
    /// for a given property will call into the associated equipment module and be handled there. 
    /// The general behavior of the property (including descriptions) can be specified with this call. 
    /// Properties can also be (and frequently are) registered via the presence of an 'exports.csv' 
    /// startup database file.
    /// Note: Properties which are overloaded can and should be registered more than once with the 
    /// appropriate characteristics supplied in the registration call.
    /// </remarks>
    /// <param name="eqm">A local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="prpName">Property name in question (up to 32 characters, preferably 16 or less).</param>
    /// <param name="dtOut">Atemplate data object giving the default format size, and structure tag 
    /// (if any) of the output data set to be returned by the server to the caller. </param>
    /// <param name="dtIn">A template data object giving the default format size, and structure tag 
    /// (if any) of the input data set to be supplied by the caller. </param>
    /// <param name="accessMode">the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server. This parameter can 
    /// be used to coerce client-side links into using preferred link attributes. For example, by applying the CA_NETWORK 
    /// flag (CA_READ|CA_NETWORK) to the access, asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the 
    /// client side. Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the registered property 
    /// (i.e. the same security rules applied to CA_WRITE will be in force for READ acquisitions, either permanently - no CA_READ flag, 
    /// or temporarily in case of an applied access lock - if both CA_READ and CA_XREAD are applied). Applying the CA_SAVERESTORE flag 
    /// will apply the internal save-and-restore logic to the registered property (restoring the last property settings upon server 
    /// startup, and saving any changed property settings).</param>
    /// <param name="arrayType">Data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the returned data.</param>
    /// <param name="rowLength">if > 0 gives the row length of the returned array data. This is most relevant for double arrays, where 
    /// the number of rows is then given by the output data size divided by the row length.</param>
    /// <param name="prpDescription">The 64-character description of what the property reads or writes. </param>
    /// <param name="prpId">The property identifier to be associated with the property name returned in calls to GetPropertyId()</param>
    /// <param name="redirectionTarget">A redirection string, if the property is to be redirected to another server. 
    /// Most properties are not redirected, and this parameter is usually NULL. If used, it will be parsed according to 
    /// device server/device name (device property), where the 'device name' and 'device property' are optional. 
    /// If 'device name' is not present, all devices will be redirected to the specified device server. If 'device property' 
    /// is not present then the identical property name as registered will be redirected to the device server specified. 
    /// If on the other hand, the device property is present in the string to be parsed, then the property registered 
    /// will be redirected to the device server specified as well as to the device property specified.</param>
    /// <returns>The (positive) associated property identifier if successful, otherwise the negative of a TINE return code.</returns>
    public static Int32 RegisterPropertyInformation(String eqm, String prpName, TDataType dtOut, TDataType dtIn, UInt16 accessMode, UInt16 arrayType, UInt16 rowLength, String prpDescription, Int32 prpId, String redirectionTarget)
    {
      DTYPE dout = new DTYPE();
      DTYPE din = new DTYPE();
      if (dtOut != null)
      {
        dout.dArrayLength = dtOut.GetDataArrayLength();
        dout.dFormat = (Int16)dtOut.GetDataFormat();
        dout.dTag = new char[Constants.TAG_NAME_SIZE];
        String dtag = dtOut.GetDataTag();
        Int32 tlen = dtag != null ? Math.Min(Constants.TAG_NAME_SIZE, dtag.Length) : 0;
        if (tlen > 0 && dtag != null)
        {
          Array.Copy(dtag.ToCharArray(), dout.dTag, (int)Math.Min(Constants.TAG_NAME_SIZE, dtag.Length));
        }
      }
      if (dtIn != null)
      {
        din.dArrayLength = dtIn.GetDataArrayLength();
        din.dFormat = (Int16)dtIn.GetDataFormat();
        din.dTag = new char[Constants.TAG_NAME_SIZE];
        String dtag = dtIn.GetDataTag();
        Int32 tlen = dtag != null ? Math.Min(Constants.TAG_NAME_SIZE, dtag.Length) : 0;
        if (tlen > 0 && dtag != null)
        {
          Array.Copy(dtag.ToCharArray(), din.dTag, (int)Math.Min(Constants.TAG_NAME_SIZE, dtag.Length));
        }
      }
      return native.RegisterPropertyInformation(eqm, prpName, ref dout, ref din, accessMode, arrayType, rowLength, prpDescription, prpId,redirectionTarget);
    }    
    /// <summary>
    /// Assigns a device name to the specified device number.
    /// </summary>
    /// <remarks>
    /// Many servers handle instances of a given device type, such as a set of vacuum pumps, 
    /// beam position monitors, magnet power supply controllers, etc. It is frequently very efficient 
    /// to code in terms of device numbers which might be an index into an array, but to specify modules 
    /// in calls by their human-readable names. For such purposes a device name can be assigned to a number 
    /// at initialization via a call to RegisterDeviceName() (alias: RegisterModuleName()). An alternative 
    /// is to provide a startup database file 'devices.csv' containing a cross-reference for numbers and names. 
    /// Internally, a hash table is maintained for fast lookups inside equipment module routines.
    /// </remarks>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name to be assigned</param>
    /// <param name="devNr">The device number associated with the device name specified.</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 RegisterDeviceName(String eqm, String devName,Int32 devNr)
    {
      return native.RegisterDeviceName(eqm, devName, devNr);
    }
    public static void SetHistoryFilesRepository(String path)
    {
      native.SetHistoryFilesRepository(path);
    }
    public static void SetAppVersion(int major, int minor, int revision)
    {
      native.SetAppVersion(major, minor, revision);
    }
    private static DateTime epoch = new DateTime(1970, 1, 1, 0, 0, 0, 0);
    public static void SetAppDate(DateTime dt)
    {
      IntPtr t=(IntPtr)0;
      TimeSpan span = (dt - epoch);
      t = (IntPtr)span.TotalSeconds;
      native.SetAppDate(t);
    }
  }
  /// <summary>
  /// Represents a TINE tagged structure.
  /// </summary>
  /// <remarks>
  /// A tagged structure is a user-defined data type that can carry any set of 
  /// TINE data types (including other tagged structures) as an atomic unit.
  /// </remarks>
  public class TTaggedStruct 
  {
    private string tag;
    public string GetTag() { return tag; }
    public void SetTag(string structTag) { tag = structTag; }
    private string className;
    public string GetClassName() { return className; }
    /// <summary>
    /// The primary constructor for a TINE tagged structure.
    /// </summary>
    /// <remarks>
    /// Using this constructor will assign the structure's 'tag' name from the
    /// .NET structure's class name and the field names from the associated field 
    /// names.
    /// But note: A structure's 'tag' name as well as the field names within a TINE tagged 
    /// structure cannot exceed 16 characters. 
    /// An alternative is to pass a null object and assign the structure's tag name and
    /// fields manually.
    /// </remarks>
    /// <param name="tstruct">is a .NET structure which will be examined and converted
    /// into a TINE tagged structure.  
    /// </param>
    /// \b Example:
    /// \include eg_RegisterStructure.cs
    public TTaggedStruct(object tstruct)
    {
      fixStructLayout(tstruct, null);
    }
    /// <summary>
    /// The primary constructor for a TINE tagged structure.
    /// </summary>
    /// <remarks>
    /// Using this constructor will assign the structure's 'tag' name according to the
    /// tag parameter passed.  If a null or empty string is passed then the constructor
    /// reverts to its simpler form using the 
    /// .NET structure's class name and the field names from the associated field 
    /// names.
    /// But note: A structure's 'tag' name as well as the field names within a TINE tagged 
    /// structure cannot exceed 16 characters. 
    /// An alternative is to pass a null object and assign the structure's tag name and
    /// fields manually.
    /// </remarks>
    /// <param name="tstruct">is a .NET structure which will be examined and converted
    /// into a TINE tagged structure.  
    /// </param>
    /// <param name="tstag">is a structure tag name to be assigned to the structure registered.
    /// </param>
    /// \b Example:
    /// \include eg_RegisterStructure.cs
    public TTaggedStruct(object tstruct, string tstag)
    {
      fixStructLayout(tstruct, tstag);
    }
    private TTaggedStruct() { }
    private static TTaggedStruct instance = null;
    public static int RegisterStructure(object tstruct,string tstag)
    {
      if (tstruct == null) return Errors.argument_list_error;
      if (instance == null) instance = new TTaggedStruct(); // create an empty one
      Type t = (tstruct is Array) ? ((Array)tstruct).GetType().GetElementType() : tstruct.GetType();
      string cn = instance.GetClassName();
      if (cn != null && cn.CompareTo(t.Name) == 0) return 0; // done (same is current instance)
      return instance.fixStructLayout(tstruct, tstag);
    }
    public static int RegisterStructure(object tstruct)
    {
      return RegisterStructure(tstruct, null);
    }
    private int fixStructLayout(object tstruct, string stag)
    {
      if (tstruct == null) return Errors.argument_list_error;
      Type stype = tstruct.GetType();
      int siz = 1;
      if (tstruct is Array)
      {
        siz = ((Array)tstruct).Length;
        stype = ((Array)tstruct).GetType().GetElementType();
      }
      className = stype.Name;
      tag = (stag != null && stag.Length > 0) ? stag : stype.Name;
      int fsiz = 1;
      Type ftype;
      FieldInfo[] fields = stype.GetFields();
      int off0, off1, fmt;
      for (int i = 0; i < fields.Length; i++)
      {
        ftype = fields[i].FieldType;
        fmt = TDataType.GetTineFormat(ftype);
        if (fields[i].FieldType.IsArray)
        {
          ftype = ftype.GetElementType();
          fmt = TDataType.GetTineFormat(ftype);
          off0 = (int)Marshal.OffsetOf(fields[i].DeclaringType, fields[i].Name);
          off1 = i < fields.Length - 1 ? ((int)Marshal.OffsetOf(fields[i + 1].DeclaringType, fields[i + 1].Name)) : Marshal.SizeOf(stype);
          fsiz = (off1 - off0) / Formats.GetSizeInBytes(fmt);
        }
        else
        {
          fsiz = 1;
        }
        AddField(fsiz, fmt, fields[i].Name);
      }
      return Seal(Marshal.SizeOf(stype), siz);
    }
    private int addrOffset = 0;
    public int AddField(Int32 siz, Int32 fmt, string field)
    {
      int cc = AddField(tag, addrOffset, siz, fmt, field);
      if (cc == 0) addrOffset += siz * Formats.GetSizeInBytes(fmt);
      return cc;
    }
    public int Seal(Int32 sizeInBytes,Int32 number)
    {
      //addrOffset = 0;
      return SealTaggedStruct(tag, sizeInBytes, number);
    }
    private static int AddField(string tag,Int32 addr,Int32 siz,Int32 fmt,string field)
    {
      fmt = Formats.LFMT_OFFSET + (fmt%256);
      return native.AddFieldToStruct(tag,addr,siz,fmt,field);
    }
    private static int SealTaggedStruct(string tag,Int32 sizeInBytes,Int32 number)
    {
      return native.SealTaggedStruct(tag, sizeInBytes, number);
    }
  }
  /// <summary>
  /// TEquipmentModule is the a TINE server's equipment module class.  
  /// </summary>
  /// <remarks>
  /// An equipment module represents a device server.  This is ideally a class which manages
  /// like instances of a set of devices.  The equipment module is responsible for
  /// handling dispatch actions pertaining to registered properties ('READ' or 'WRITE'
  /// actions). 
  /// </remarks>
  public class TEquipmentModule
  {
    private CallHandler eqmHndlr;
    private native.TEqmFcn eqmFcn;
    private FcnDispatch eqmIniDispatch;
    private native.TEqmHook eqmIni;
    private FcnDispatch eqmBkgDispatch;
    private native.TEqmHook eqmBkg;
    private FcnDispatch eqmExiDispatch;
    private native.TEqmHook eqmExi;
    private string eqmName;
    public unsafe Int32 lclHndlr(string dev, string prp, ref DTYPE dout, ref DTYPE din, UInt16 acc, IntPtr eqmRef)
    {
      TDataType dtdout = new TDataType(dout);
      TDataType dtdin = new TDataType(din);
      Int32 cc = eqmHndlr(dev, prp, dtdout, dtdin, acc, this);
      dout.dArrayLength = dtdout.GetDataArrayLength();
      dout.dTimeStamp = dtdout.GetDataTimeStamp();
      dout.dStamp = dtdout.GetUserStamp();
      dout.sysStamp = dtdout.GetSystemStamp();
      return cc;
    }
    public unsafe void lclIni(IntPtr eqmRef)
    {
      eqmIniDispatch();
    }
    public unsafe void lclBkg(IntPtr eqmRef)
    {
      eqmBkgDispatch();
    }
    public unsafe void lclExi(IntPtr eqmRef)
    {
      eqmExiDispatch();
    }
    /// <summary>
    /// Primary constructor for an equipment module object.
    /// </summary>
    /// <remarks>
    /// A FEC (front-end controller) process can contain more than one equipment module (which is exported
    /// as a device server or property server).  However, a FEC process should contain at least one 
    /// equipment module in order to be a useful element in the control system.
    /// </remarks>
    /// <param name="exp">is the equipment module's exported device or property server name 
    /// (maximum 32 characters).  This name must be unique within a given context.
    /// If a 'null' parameter is passed, then the exported name must be given by an underlying 
    /// configuration file such as 'export.csv' or 'fec.xml'</param>
    /// <param name="eqm">is the equipment module's internal 'local' name (maximum 6-characters).
    /// This name must be unique within the process only, and is used for internal de-referencing.
    /// </param>
    /// <param name="ndev">gives the maximum device capacity (maximum number of device instances
    /// supported by this device server.</param>
    /// <param name="hndlr">is the equipment module's dispatch call handler.  This must be a function
    /// of the form:
    /// int CallHandler(string dev, string prp, TDataType dout, TDataType din, UInt16 acc, TEquipmentModule eqm);
    /// </param>
    /// <param name="ini">is an optional initialization routine for the equipment module.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// <param name="bkg">is an optional background routine for the equipment module, which will be called
    /// at the given background task interval.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// <param name="bkgInterval">is the background task interval in play if a background task is registered.</param>
    /// <param name="exi">is an optional background routine for the equipment module, which will be called
    /// at the given background task interval.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// \b Example:
    /// \include eg_RegisterEqm.cs
    public TEquipmentModule(string exp, string eqm, Int32 ndev, CallHandler hndlr, FcnDispatch ini, FcnDispatch bkg, Int32 bkgInterval, FcnDispatch exi)
    {
      eqmHndlr = hndlr;
      eqmFcn = lclHndlr;
      eqmIniDispatch = ini;
      eqmIni = lclIni;
      eqmBkgDispatch = bkg;
      eqmBkg = lclBkg;
      eqmExiDispatch = exi;
      eqmExi = lclExi;
      eqmName = (string)eqm.Clone();
      native.RegisterEquipmentModuleEx(exp, eqmName, (UInt32)ndev, eqmFcn, eqmIni, eqmBkg, bkgInterval, eqmExi, IntPtr.Zero);
    }
    /// <summary>
    /// Assigns pertinent information for the specified property. 
    /// </summary>
    /// <remarks>
    /// Servers export their behavior through Properties, which can be thought of as 'properties' 
    /// in the Object Oriented sense with Get and Set methods. That is, any read or write access 
    /// for a given property will call into the associated equipment module and be handled there. 
    /// The general behavior of the property (including descriptions) can be specified with this call. 
    /// Properties can also be (and frequently are) registered via the presence of an 'exports.csv' 
    /// startup database file.
    /// Note: Properties which are overloaded can and should be registered more than once with the 
    /// appropriate characteristics supplied in the registration call.
    /// </remarks>
    /// <param name="prpName">Property name in question (up to 32 characters, preferably 16 or less).</param>
    /// <param name="dtOut">Atemplate data object giving the default format size, and structure tag 
    /// (if any) of the output data set to be returned by the server to the caller. </param>
    /// <param name="dtIn">A template data object giving the default format size, and structure tag 
    /// (if any) of the input data set to be supplied by the caller. </param>
    /// <param name="accessMode">the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server. This parameter can 
    /// be used to coerce client-side links into using preferred link attributes. For example, by applying the CA_NETWORK 
    /// flag (CA_READ|CA_NETWORK) to the access, asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the 
    /// client side. Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the registered property 
    /// (i.e. the same security rules applied to CA_WRITE will be in force for READ acquisitions, either permanently - no CA_READ flag, 
    /// or temporarily in case of an applied access lock - if both CA_READ and CA_XREAD are applied). Applying the CA_SAVERESTORE flag 
    /// will apply the internal save-and-restore logic to the registered property (restoring the last property settings upon server 
    /// startup, and saving any changed property settings).</param>
    /// <param name="arrayType">Data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the returned data.</param>
    /// <param name="rowLength">if > 0 gives the row length of the returned array data. This is most relevant for double arrays, where 
    /// the number of rows is then given by the output data size divided by the row length.</param>
    /// <param name="prpDescription">The 64-character description of what the property reads or writes. </param>
    /// <param name="prpId">The property identifier to be associated with the property name returned in calls to GetPropertyId()</param>
    /// <param name="redirectionTarget">A redirection string, if the property is to be redirected to another server. 
    /// Most properties are not redirected, and this parameter is usually NULL. If used, it will be parsed according to 
    /// device server/device name (device property), where the 'device name' and 'device property' are optional. 
    /// If 'device name' is not present, all devices will be redirected to the specified device server. If 'device property' 
    /// is not present then the identical property name as registered will be redirected to the device server specified. 
    /// If on the other hand, the device property is present in the string to be parsed, then the property registered 
    /// will be redirected to the device server specified as well as to the device property specified.</param>
    /// <returns>The (positive) associated property identifier if successful, otherwise the negative of a TINE return code.</returns>
    /// \b Example:
    /// \include eg_RegisterProperty.cs
    /// 
    public int RegisterPropertyInformation(string property, TDataType dout, TDataType din, UInt16 acc, UInt16 atype, int rowLength, string dsc, int propId, string rdr)
    {
      return TKernel.RegisterPropertyInformation(eqmName, property, dout, din, acc, atype, (ushort)rowLength, dsc, propId, rdr);
    }
    /// <summary>
    /// Assigns a device name to the specified device number.
    /// </summary>
    /// <remarks>
    /// Many servers handle instances of a given device type, such as a set of vacuum pumps, 
    /// beam position monitors, magnet power supply controllers, etc. It is frequently very efficient 
    /// to code in terms of device numbers which might be an index into an array, but to specify modules 
    /// in calls by their human-readable names. For such purposes a device name can be assigned to a number 
    /// at initialization via a call to RegisterDeviceName() (alias: RegisterModuleName()). An alternative 
    /// is to provide a startup database file 'devices.csv' containing a cross-reference for numbers and names. 
    /// Internally, a hash table is maintained for fast lookups inside equipment module routines.
    /// </remarks>
    /// <param name="devName">Up to 64 characters device name to be assigned</param>
    /// <param name="devNr">The device number associated with the device name specified.</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    /// \b Example:
    /// \include eg_RegisterDevice.cs
    public int RegisterDeviceName(String devName, Int32 devNr)
    {
      return TKernel.RegisterDeviceName(eqmName, devName, devNr);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device number. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <returns></returns>
    public int SetAlarm(Int32 devNr, Int32 almCode)
    {
      return native.SetAlarmEx(eqmName, (Int16)devNr, almCode, IntPtr.Zero, (byte)0);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device number. 
    /// This methods also allows passing associated alarm data and specific alarm flags.
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <param name="almData">is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.</param>
    /// <param name="almFlags">is the optional set of alarm flags which is used in the alarm descriptor information.</param>
    /// <returns></returns>
    public int SetAlarm(Int32 devNr, Int32 almCode, object almData, byte almFlags)
    {
      IntPtr ptr;
      if (almData is string)
      {
        char[] cdata = ((string)almData).ToCharArray();
        byte[] bdata = new byte[64];
        for (int i = 0; i < 64 && i < cdata.Length; i++) bdata[i] = (byte)cdata[i];
        ptr = TDataType.GetDataPtr(bdata);
      }
      else
      {
        ptr = TDataType.GetDataPtr(almData);
      }
      return native.SetAlarmEx(eqmName, (Int16)devNr, almCode, ptr, almFlags);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device name
    /// (as opposed to its device number).
    /// This methods also allows passing associated alarm data and specific alarm flags.
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName">is the device name for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <param name="almData">is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.</param>
    /// <param name="almFlags">is the optional set of alarm flags which is used in the alarm descriptor information.</param>
    /// <returns></returns>
    public int SetAlarm(string devName, Int32 almCode, object almData, byte almFlags)
    {
      IntPtr ptr;
      if (almData is string)
      {
        char[] cdata = ((string)almData).ToCharArray();
        byte[] bdata = new byte[64];
        for (int i = 0; i < 64 && i < cdata.Length; i++) bdata[i] = (byte)cdata[i];
        ptr = TDataType.GetDataPtr(bdata);
      }
      else
      {
        ptr = TDataType.GetDataPtr(almData);
      }
      return native.SetDeviceAlarm(eqmName, devName, almCode, ptr, almFlags);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be cleared. 
    /// </summary>
    /// <remarks>
    /// A server can clear alarms for specific device by giving its device number. 
    /// Note that all alarms for the device specified are cleared. 
    /// Also note, that "clearing" an alarm means that its "clear" counter is incremented. 
    /// It is not removed immediately from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which all alarms is to be cleared.</param>
    /// 
    public void ClearAlarm(Int32 devNr)
    {
      native.ClearAlarm(eqmName,(Int16)devNr);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be cleared. 
    /// </summary>
    /// <remarks>
    /// A server can clear alarms for specific device by giving its device name. 
    /// Note that all alarms for the device specified are cleared. 
    /// Also note, that "clearing" an alarm means that its "clear" counter is incremented. 
    /// It is not removed immediately from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName">is the registered device name of the device for which all alarms is to be cleared.</param>
    public void ClearAlarm(string devName)
    {
      native.ClearDeviceAlarm(eqmName, devName);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be marked for removal.
    /// </summary>
    /// <remarks>
    /// A server can remove alarms for specific device by giving its registered device name. 
    /// Note that only the alarm specified for the device specified is removed. 
    /// Also note, that "removing" an alarm means that termination bit will be set. 
    /// It is subsequently marked for removal from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName"></param>
    /// <param name="almCode"></param>
    public void RemoveAlarm(string devName,Int32 almCode)
    {
      native.RemoveDeviceAlarm(eqmName, devName, almCode);
    }
    /// <summary>
    /// Inserts an alarm definition into the alarm definition table.
    /// </summary>
    /// <remarks>
    /// As an alternative to the <local name>-alarms.csv configuration file, the front end 
    /// server can make use of this API call in order to fill in the alarm definition table 
    /// describing locally generated alarms. This is particularly useful for embedded platforms 
    /// where there is no file system, or where a TINE server is used as a translation layer and 
    /// needs to map a given alarm system onto the TINE alarm system.
    /// </remarks>
    /// <param name="ads">is a pointer to an Alarm Definition Structure (ADS) containing the alarm table information which is to be appended to the alarm definition table.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    public int AppendAlarmInfoTable(ref ADS ads)
    {
      return native.AppendAlarmInfoTable(eqmName,ref ads);
    }
    /// <summary>
    /// Inserts a property to be monitored into the local alarm server's Watch Table.
    /// </summary>
    /// <remarks>
    /// Certain alarms are to be set whenever the value of a property exceeds a defineable threshold. 
    /// Such alarms can be managed automatically by the local alarm server if the alarm criteria are 
    /// entered into the alarm watch table. 
    /// This can be achieved by calling this routine (or supplying a startup configuration file 
    /// almwatch.csv).
    /// </remarks>
    /// <param name="prp">is the property which is to be called by the local alarm server.</param>
    /// <param name="dev">is the device name associated with the property to be called by the local alarm server.</param>
    /// <param name="siz">is the data array size to be called by the local alarm server.</param>
    /// <param name="fmt">is the TINE data format to be called by the local alarm server.</param>
    /// <param name="atyp">is the TINE data array type to be applied to the property called by the local alarm server.</param>
    /// <param name="sev">is the severity of the alarm issued when the data returned by the call exceed the given thresholds.</param>
    /// <param name="sys">is the alarm system identifier to be associated with the alarm.</param>
    /// <param name="thr">is an ALM_THRESHOLDS object specifying the threshold criteria for setting the alarm.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    public int AppendAlarmWatchTable(string prp, string dev, Int32 siz, Int32 fmt, Int32 atyp, Int32 sev, Int32 sys, ref ALM_THRESHOLDS thr)
    {
      return native.AppendAlarmWatchTable(eqmName, prp, dev, siz, fmt, atyp, sev, sys, ref thr);
    }
    /// <summary>
    /// Inserts a local history element into the local history server. 
    /// </summary>
    /// <remarks>
    /// A server can instruct the local history server to keep a history of the given property by utilizing this call. The local history server will periodically call the property as specified according to the following input parameters.
    ///</remarks> 
    /// <param name="prp">is the requested property for which a history is to be kept.</param>
    /// <param name="dev">is the device name to be associated with the property name supplied as the second parameter.</param>
    /// <param name="len">is the length of the local history call.</param>
    /// <param name="fmt">is the TINE format of the local history call.</param>
    /// <param name="idx">is the local history index to be identified with this local history element (Note: this must be unique with this server process and within the history data repository).</param>
    /// <param name="hspec">is a pointer to a HistorySpecification structure, where the relevant filtering criteria are given.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    /// \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the startup configuration file 'history.csv', in which all local history critieria can be introduced. This has the advantage that no information is hard-coded. Indeed by introducing a 'history.csv' file one can immediately instruct a server to maintain local history information (even if the server is developed and maintained by someone else!).
    /// 
    /// \b Example:
    /// \include eg_AppendHistory.cs
    public int AppendHistoryInformation(string prp, string dev, Int32 len, Int32 fmt, Int32 idx, ref HistorySpecification hspec)
    {
      return native.AppendHistoryInformation(eqmName, prp, dev, len, fmt, idx, ref hspec);
    }
    public delegate int CallHandler(string dev, string prp, TDataType dout, TDataType din, UInt16 acc, TEquipmentModule eqm);
    public delegate void FcnDispatch();
    /// <summary>
    /// Returns the registered device number for the device name given.
    /// </summary>
    /// <param name="device">is the device name whose assigned device number is desired.
    /// If the device name entered does not correspond to a registered device the value
    /// '-1' is returned.
    /// </param>
    /// <returns>The corresponding device number or '-1' on error.</returns>
    public Int32 GetDeviceNumber(string device) { return native.GetDeviceNumber(eqmName, device); }
    /// <summary>
    /// Returns the registered device number for the device name and property given.
    /// </summary>
    /// <remarks>
    /// A pure device server will maintain a list of registered devices, each of which may support
    /// any particular given property.  In such cases the 'property' parameter is irrelevant, as any
    /// associated device number is independent of the property.
    /// 
    /// On the other hand, a 'property' server will treat properties as the more relevant identifier,
    /// whereby each property may support an independent set of 'device' names (or more probably 
    /// 'keyword' names).  In such cases the property name entered is of paramount importance in
    /// determining the registered device (or keyword) number.
    /// </remarks>
    /// <param name="device">is the device name whose assigned device number is desired.
    /// If the device name entered does not correspond to a registered device the value
    /// '-1' is returned.
    /// </param>
    /// <param name="property">is the property whose device set is being queried.</param>
    /// <returns>The corresponding device number or '-1' on error.</returns>
    /// \b Example:
    /// \include eg_GetPropId.cs
    public Int32 GetDeviceNumber(string device,string property) { return native.GetDeviceNumberEx(eqmName, device, property); }
    /// <summary>
    /// Returns the registered property identification number for the property entered.
    /// </summary>
    /// <param name="property">is the property whose assigned property identification is desired</param>
    /// <returns>The property identication number of -1 on error.</returns>
    public Int32 GetPropertyId(string property) { return native.GetPropertyId(eqmName, property); }
    public unsafe Int32 GetCallerInformation(ClnInfo[] clnInfo, ref int num) 
    { 
     // return native.GetCallerInformation(eqmName, clnInfo, ref num); 
      return 0;
    }
    public unsafe ClnInfo GetCallerInformation()
    {
      int num = 1;
      ClnInfo ci = new ClnInfo();
      int cc = native.GetCallerInformation(eqmName, ref ci, ref num);
      return ci;
    }
    /// <summary>
    /// Schedules the given property for immediate delivery to all attached clients according scope specified.
    /// </summary>
    /// <remarks>
    /// When clients are attached to a particular property, they have specified a polling rate, which 
    /// defines the maximum latency for receiving updates at the client side. This is usually fine. 
    /// However, if a server knows that important data have changed and which properties depend on these 
    /// data, it can signal the scheduler to call the given properties immediately and send the data to 
    /// all attached callers, regardless of their registered polling rates. In this way, a server can 
    /// signal an event to its attached clients. In this extended call, the caller can specify the 
    /// 'scope' of the call : CA_NETWORK (all listening clients), CA_HIST (the local history subsystem) 
    /// or CA_ALARM (the local alarm server).
    /// </remarks>
    /// <param name="property">is (comma separated) list of properties which are to be scheduled.</param>
    /// <param name="scope">is any of CA_NETWORK, CA_HIST, or CA_ALARM ORed together.</param>
    /// <returns>0 upon success or a TINE error code.</returns>
    public Int32 ScheduleProperty(string property,Int32 scope) { return native._SystemSchedulePropertyEx(eqmName, property, scope); }
    /// <summary>
    /// Schedules the given property for immediate delivery to all attached clients.
    /// </summary>
    /// <remarks>
    /// When clients are attached to a particular property, they have specified a polling rate, which 
    /// defines the maximum latency for receiving updates at the client side. This is usually fine. 
    /// However, if a server knows that important data have changed and which properties depend on these 
    /// data, it can signal the scheduler to call the given properties immediately and send the data to 
    /// all attached callers, regardless of their registered polling rates. In this way, a server can 
    /// signal an event to its attached clients. In this simple call, clients from all listening
    /// subsystems will be scheduled (i.e. the local history and alarm subsystem as well as any remote
    /// clients).
    /// </remarks>
    /// <param name="property">is (comma separated) list of properties which are to be scheduled.</param>
    /// <returns></returns>
    public Int32 ScheduleProperty(string property) 
    { 
      return native._SystemSchedulePropertyEx(eqmName, property, Access.CA_NETWORK | Access.CA_HIST | Access.CA_ALARM); 
    }
    /// <summary>
    /// Sets the error string to accompany the current server call
    /// </summary>
    /// <remarks>
    ///  * For user-defined errors codes (equal to or above 512) the accompanying error string
    /// should be defined by calling this routine prior to returning from the equipment module
    /// call.  When system error codes such as 'illegal_format' are used, a call to SetEqmCompletion() 
    /// is not necessary.
    /// </remarks>
    /// <param name="statusString">
    /// is the 96-character string which is to accompany the outgoing error code.
    /// </param>
    public void SetCompletion(string statusString)
    {
      native.SetEqmCompletion(eqmName, statusString);
    }
  }
}
